import struct
import numpy as np
from astropy import units as u
from astropy.coordinates import Angle
import os

###
# Header parsing
###

# Dictionary of allowed keywords and their types
# Here are the keywordss that a filter bank file may
# contain.  Items marked with "[*]" are not yet # supported.  See docs for
# indivisuabl attribtues for more detailed info.
#
#   * telescope_id (int): 0=fake data; 1=Arecibo; 2=Ooty... others to be added
#   * machine_id (int): 0=FAKE; 1=PSPM; 2=WAPP; 3=OOTY... others to be added
#   * data_type (int): 1=blimpy; 2=time series... others to be added
#   * rawdatafile (string): the name of the original data file
#   * source_name (string): the name of the source being observed by the telescope
#   * barycentric (int): equals 1 if data are barycentric or 0 otherwise
#   * pulsarcentric (int): equals 1 if data are pulsarcentric or 0 otherwise
#   * az_start (double): telescope azimuth at start of scan (degrees)
#   * za_start (double): telescope zenith angle at start of scan (degrees)
#   * src_raj (double): right ascension (J2000) of source (hours, converted from hhmmss.s)
#   * src_dej (double): declination (J2000) of source (degrees, converted from ddmmss.s)
#   * tstart (double): time stamp (MJD) of first sample
#   * tsamp (double): time interval between samples (s)
#   * nbits (int): number of bits per time sample
#   * nsamples (int): number of time samples in the data file (rarely used any more)
#   * fch1 (double): centre frequency (MHz) of first blimpy channel
#   * foff (double): blimpy channel bandwidth (MHz)
#   * FREQUENCY_START [*] (character): start of frequency table (see below for explanation)
#   * fchannel [*] (double): frequency channel value (MHz)
#   * FREQUENCY_END [*] (character): end of frequency table (see below for explanation)
#   * nchans (int): number of blimpy channels
#   * nifs (int): number of seperate IF channels
#   * refdm (double): reference dispersion measure (pc/cm**3)
#   * period (double): folding period (s)
#   * nbeams (int):total number of beams (?)
#   * ibeam (int): number of the beam in this file (?)

header_keyword_types = {
    'telescope_id' : '<l',
    'machine_id'   : '<l',
    'data_type'    : '<l',
    'barycentric'  : '<l',
    'pulsarcentric': '<l',
    'nbits'        : '<l',
    'nsamples'     : '<l',
    'nchans'       : '<l',
    'nifs'         : '<l',
    'nbeams'       : '<l',
    'ibeam'        : '<l',
    'rawdatafile'  : 'str',
    'source_name'  : 'str',
    'az_start'     : '<d',
    'za_start'     : '<d',
    'tstart'       : '<d',
    'tsamp'        : '<d',
    'fch1'         : '<d',
    'foff'         : '<d',
    'refdm'        : '<d',
    'period'       : '<d',
    'src_raj'      : 'angle',
    'src_dej'      : 'angle',
}


def len_header(filename):
    """ Return the length of the blimpy header, in bytes

    Args:
        filename (str): name of file to open

    Returns:
        idx_end (int): length of header, in bytes
    """
    GULP = 4096
    with open(filename, 'rb') as f:
        header_sub_count = 0
        eoh_found = False
        while not eoh_found:
            header_sub = f.read(GULP)
            header_sub_count += 1
            if b'HEADER_END' in header_sub:
                idx_end = header_sub.index(b'HEADER_END') + len(b'HEADER_END')
                eoh_found = True
                break

        idx_end = (header_sub_count -1) * GULP + idx_end
    return idx_end


def read_next_header_keyword(fh):
    """

    Args:
        fh (file): file handler

    Returns:
    """
    n_bytes = np.frombuffer(fh.read(4), dtype='uint32')[0]

    if n_bytes > 255:
        n_bytes = 16

    keyword = fh.read(n_bytes).decode('ascii')

    if keyword in ('HEADER_START', 'HEADER_END'):
        return keyword, 0, fh.tell()
    dtype = header_keyword_types[keyword]
    idx = fh.tell()
    if dtype == '<l':
        val = struct.unpack(dtype, fh.read(4))[0]
    if dtype == '<d':
        val = struct.unpack(dtype, fh.read(8))[0]
    if dtype == 'str':
        str_len = np.frombuffer(fh.read(4), dtype='uint32')[0]
        val = fh.read(str_len).decode('ascii')
    if dtype == 'angle':
        val = struct.unpack('<d', fh.read(8))[0]
        val = fil_double_to_angle(val)
        if keyword == 'src_raj':
            val = Angle(val, unit=u.hour)
        else:
            val = Angle(val, unit=u.deg)
    return keyword, val, idx


def is_filterbank(filename):
    """ Open file and confirm if it is a filterbank file or not. """
    with open(filename, 'rb') as fh:
        is_fil = True

        # Check this is a blimpy file
        try:
            keyword, value, idx = read_next_header_keyword(fh)
            try:
                assert keyword == 'HEADER_START'
            except AssertionError:
                is_fil = False
        except KeyError:
            is_fil = False
        return is_fil


def read_header(filename, return_idxs=False):
    """ Read blimpy header and return a Python dictionary of key:value pairs

    Args:
        filename (str): name of file to open

    Optional args:
        return_idxs (bool): Default False. If true, returns the file offset indexes
                            for values

    returns

    """
    with open(filename, 'rb') as fh:
        header_dict = {}
        header_idxs = {}

        # Check this is a blimpy file
        keyword, value, idx = read_next_header_keyword(fh)

        try:
            assert keyword == 'HEADER_START'
        except AssertionError:
            raise RuntimeError("Not a valid blimpy file.")

        while True:
            keyword, value, idx = read_next_header_keyword(fh)
            if keyword == 'HEADER_END':
                break
            header_dict[keyword] = value
            header_idxs[keyword] = idx

    if return_idxs:
        return header_idxs
    return header_dict

def fix_header(filename, keyword, new_value):
    """ Apply a quick patch-up to a Filterbank header by overwriting a header value


    Args:
        filename (str): name of file to open and fix. WILL BE MODIFIED.
        keyword (stt):  header keyword to update
        new_value (long, double, angle or string): New value to write.

    Notes:
        This will overwrite the current value of the blimpy with a desired
        'fixed' version. Note that this has limited support for patching
        string-type values - if the length of the string changes, all hell will
        break loose.

    """

    # Read header data and return indexes of data offsets in file
    hd = read_header(filename)
    hi = read_header(filename, return_idxs=True)
    idx = hi[keyword]

    # Find out the datatype for the given keyword
    dtype = header_keyword_types[keyword]
    dtype_to_type = {'<l'  : np.int32,
                     'str' : bytes,
                     '<d'  : np.float64,
                     'angle' : to_sigproc_angle}
    value_dtype = dtype_to_type[dtype]

    # Generate the new string
    if isinstance(value_dtype, bytes):
        if len(hd[keyword]) == len(new_value):
            val_str = np.int32(len(new_value)).tostring() + new_value
        else:
            raise RuntimeError("String size mismatch. Cannot update without rewriting entire file.")
    else:
        val_str = value_dtype(new_value).tostring()

    # Write the new string to file
    with open(filename, 'rb+') as fh:
        fh.seek(idx)
        fh.write(val_str)

def fil_double_to_angle(angle):
    """ Reads a little-endian double in ddmmss.s (or hhmmss.s) format and then
    converts to Float degrees (or hours).  This is primarily used to read
    src_raj and src_dej header values. """

    negative = (angle < 0.0)
    angle = np.abs(angle)

    dd = np.floor((angle / 10000))
    angle -= 10000 * dd
    mm = np.floor((angle / 100))
    ss = angle - 100 * mm
    dd += mm/60.0 + ss/3600.0

    if negative:
        dd *= -1

    return dd

###
# sigproc writing functions
###

def to_sigproc_keyword(keyword, value=None):
    """ Generate a serialized string for a sigproc keyword:value pair

    If value=None, just the keyword will be written with no payload.
    Data type is inferred by keyword name (via a lookup table)

    Args:
        keyword (str): Keyword to write
        value (None, float, str, double or angle): value to write to file

    Returns:
        value_str (str): serialized string to write to file.
    """
    if value is None:
        return np.int32(len(keyword)).tobytes() + keyword.encode('ascii')
    dtype = header_keyword_types[keyword]

    dtype_to_type = {'<l'  : np.int32,
                     'str' : str,
                     '<d'  : np.float64,
                     'angle' : to_sigproc_angle}

    value_dtype = dtype_to_type[dtype]

    if isinstance(value, str):
        value = value.encode('ascii')
    if value_dtype is str:
        return np.int32(len(keyword)).tobytes() + keyword.encode('ascii') + np.int32(len(value)).tobytes() + value
    return np.int32(len(keyword)).tobytes() + keyword.encode('ascii') + value_dtype(value).tobytes()

def generate_sigproc_header(f):
    """ Generate a serialzed sigproc header which can be written to disk.

    Args:
        f (Filterbank object): Filterbank object for which to generate header

    Returns:
        header_str (str): Serialized string corresponding to header
    """

    header_string = b''
    header_string += to_sigproc_keyword('HEADER_START')

    for keyword in f.header.keys():
        if keyword == 'src_raj':
            header_string += to_sigproc_keyword('src_raj')  + to_sigproc_angle(f.header['src_raj'])
        elif keyword == 'src_dej':
            header_string += to_sigproc_keyword('src_dej')  + to_sigproc_angle(f.header['src_dej'])
        elif keyword in ('az_start', 'za_start'):
            header_string += to_sigproc_keyword(keyword)  + np.float64(f.header[keyword]).tobytes()
        elif keyword not in header_keyword_types.keys():
            pass
        else:
            header_string += to_sigproc_keyword(keyword, f.header[keyword])

    header_string += to_sigproc_keyword('HEADER_END')
    return header_string


def to_sigproc_angle(angle_val):
    """ Convert an astropy.Angle to the ridiculous sigproc angle format string. """
    x = str(angle_val)

    if '.' in x:
        if 'h' in x:
            d, m, s, ss = int(x[0:x.index('h')]), int(x[x.index('h')+1:x.index('m')]), \
            int(x[x.index('m')+1:x.index('.')]), float(x[x.index('.'):x.index('s')])
        if 'd' in x:
            d, m, s, ss = int(x[0:x.index('d')]), int(x[x.index('d')+1:x.index('m')]), \
            int(x[x.index('m')+1:x.index('.')]), float(x[x.index('.'):x.index('s')])
    else:
        if 'h' in x:
            d, m, s = int(x[0:x.index('h')]), int(x[x.index('h')+1:x.index('m')]), \
            int(x[x.index('m')+1:x.index('s')])
        if 'd' in x:
            d, m, s = int(x[0:x.index('d')]), int(x[x.index('d')+1:x.index('m')]), \
            int(x[x.index('m')+1:x.index('s')])
        ss = 0
    num = str(d).zfill(2) + str(m).zfill(2) + str(s).zfill(2)+ '.' + str(ss).split(".")[-1]
    return np.float64(num).tobytes()


def calc_n_ints_in_file(filename):
    """ Calculate number of integrations in a given file """
    # Load binary data
    h = read_header(filename)
    n_chans = h['nchans']
    n_ifs   = h['nifs']
    idx_data = len_header(filename)
    filesize = os.path.getsize(filename)
    n_bytes_data = filesize - idx_data

    if h['nbits'] == 2:
        n_ints = int(4 * n_bytes_data / (n_chans * n_ifs))
    elif h['nbits'] == 4:
        n_ints = int(2 * n_bytes_data / (n_chans * n_ifs))
    else:
        n_bytes  = int(h['nbits'] / 8)
        n_ints = int(n_bytes_data / (n_bytes * n_chans * n_ifs))

    return n_ints



from os.path import dirname
import numpy as np
from astropy import units as u
import setigen as stg
import matplotlib.pyplot as plt
from blimpy.signal_processing.dedoppler import dedoppler_1
from blimpy import Waterfall
from blimpy.stax import plot_waterfall, sort2
from tests.data import voyager_fil


PLOT_DIR = dirname(voyager_fil)
FIL_FILE = PLOT_DIR + "/test_dedoppler.fil"
PNG_FILE = PLOT_DIR + "/test_dedoppler.png"


# Plotting constants
fontsize = 16
font_dict = {"family" : "DejaVu Sans", "size" : fontsize}
N_PLOTS = 6


def plotter(counter, drift_rate):
    wf = Waterfall(FIL_FILE)
    dedoppler_1(wf, drift_rate)
    wf.header["source_name"] = "Dedoppler at D.R. {} Hz".format(drift_rate)
    plt.subplot(N_PLOTS, 1, counter)
    return plot_waterfall(wf)


def test_dedoppler_1():
    
    # Generate the Filterbank file.
    print("test_dedoppler_1: Creating Filterbank file {}".format(FIL_FILE))
    frame = stg.Frame(fchans=1024*u.pixel,
                      tchans=32*u.pixel,
                      df=2.7939677238464355*u.Hz,
                      dt=18.253611008*u.s,
                      fch1=6095.214842353016*u.MHz,
                      ascending=True)
    frame.add_noise(x_mean=10, noise_type='chi2')
    frame.add_signal(stg.constant_path(f_start=frame.get_frequency(index=200),
                                       drift_rate=2*u.Hz/u.s),
                                       stg.constant_t_profile(level=frame.get_intensity(snr=30)),
                                       stg.gaussian_f_profile(width=40*u.Hz),
                                       stg.constant_bp_profile(level=1))    
    frame.save_fil(FIL_FILE)
    
    # Load Filterban file.
    print("test_dedoppler_1: Loading Filterbank file {}".format(FIL_FILE))
    wf = Waterfall(FIL_FILE)
    freqs = wf.get_freqs()
    the_lowest, the_highest = sort2(freqs[0], freqs[-1])
    the_midpoint = np.abs(the_lowest + the_highest) / 2.

    # Initialise plotting.
    print("test_dedoppler_1: Plotting to file {}".format(PNG_FILE))
    fig = plt.subplots(N_PLOTS, sharex=True, sharey=True, figsize=(10, 2 * N_PLOTS))
    wf.header["source_name"] = "Initial Data"

    # Plot 1.
    plt.subplot(N_PLOTS, 1, 1)
    this_plot = plot_waterfall(wf)
    
    # Plot #2.
    this_plot = plotter(2, 1.0)
    
    # Plot #3.
    this_plot = plotter(3, 1.5)
    
    # Plot #4.
    this_plot = plotter(4, 2.2)
    
    # Plot #5.
    this_plot = plotter(5, 2.7)
    
    # Plot #6.
    this_plot = plotter(6, 3.6)
    
    # Finish up plots.
    plt.xticks(np.linspace(the_lowest, the_highest, num=4), ["","","",""])
    factor = 1e6
    units = "Hz"
    xloc = np.linspace(the_lowest, the_highest, 5)
    xticks = [round(loc_freq) for loc_freq in (xloc - the_midpoint) * factor]
    if np.max(xticks) > 1000:
        xticks = [xt / 1000 for xt in xticks]
        units = "kHz"
    plt.xticks(xloc, xticks)
    plt.xlabel("Relative Frequency [%s] from %f MHz" % (units, the_midpoint), fontdict=font_dict)
    cax = fig[0].add_axes([0.94, 0.11, 0.03, 0.77])
    fig[0].colorbar(this_plot, cax=cax, label="Normalized Power (Arbitrary Units)")
    plt.subplots_adjust(hspace=0, wspace=0)

    plt.savefig(PNG_FILE, dpi=200, bbox_inches="tight")
    print("test_dedoppler_1: End")


if __name__ == "__main__":
    test_dedoppler_1()



r"""
observatory.py
blimpy
"""

import os
import pandas as pd

DEBUGGING = False


class Observatory:
    r""" Class for handling observatory data.
        The Pandas dataframe is defined in ./observatory_info.csv
    """


    def __init__(self, telescope_id=None, telescope_name=None):
        """ init method for Observatory class
        Parameters (one or the other or neither):
            telescope_id (int): sigproc telescope_id
            telescope_name (str): telescope name
        If neither parameter is specified, then the fake telescope is used.
        Returns:
            self
        """
        abs_path = os.path.dirname(os.path.realpath(__file__))
        path_csv = os.path.join(abs_path, 'observatory_info.csv')
        df_full = pd.read_csv(path_csv, sep=',', engine='python', comment='#')
        if DEBUGGING:
            print("\nObservatory __init__ path_csv:", path_csv)
            print("Observatory __init__ df_full:\n", df_full)
        if telescope_id is not None:
            try:
                fields_dict = df_full.loc[df_full['TELESCOPE_ID'] == telescope_id].to_dict('list')
            except Exception as exc:
                raise RuntimeError("Observatory __init__ telescope_id={} is invalid!"
                                   .format(telescope_id)) from exc
        elif telescope_name is not None:
            try:
                fields_dict = df_full.loc[df_full['TELESCOPE_NAME'] == telescope_name].to_dict('list')
            except Exception as exc:
                raise RuntimeError("Observatory __init__ telescope_name={} is invalid!"
                                   .format(telescope_name)) from exc
        else:
            try:
                fields_dict = df_full.loc[df_full['TELESCOPE_ID'] == 0].to_dict('list')
            except Exception as exc:
                raise RuntimeError("Observatory __init__ Cannot find the fake telescope!") from exc
        if DEBUGGING:
            print("\nObservatory __init__ fields_dict:\n", fields_dict)
        self.telescope_name = fields_dict.get('TELESCOPE_NAME')[0]
        self.telescope_name_short = fields_dict.get('TELESCOPE_NAME_SHORT')[0]
        self.telescope_id = fields_dict.get('TELESCOPE_ID')[0]
        self.dish_diameter = fields_dict.get('DISH_DIAMETER')[0]
        self.xyz_coords = [
            fields_dict.get('X')[0],
            fields_dict.get('Y')[0],
            fields_dict.get('Z')[0] ]


    def get_telescope_name(self):
        r""" Return the telescope name to caller. """
        return self.telescope_name


    def get_telescope_name_short(self):
        r""" Return the short telescope name to caller. """
        return self.telescope_name_short


    def get_telescope_id(self):
        r""" Return the SIGPROC ID to caller. """
        return self.telescope_id


    def get_xyz_coords(self):
        r""" Return the X Y Z coordinates to caller. """
        return self.xyz_coords


    def get_dish_diameter(self):
        r""" Return the dish diameter to caller. """
        return self.dish_diameter


    def calc_beam_halfwidth(self, freq):
        """ Calculates beam halfwidth
            Code adapted from PRESTO
            Note: returns -1 if dish diameter = -1.
        Parameters:
            freq (int or float): frequency in MHz
        Returns:
            float: beam halfwidth in arcsec
        """
        # constants from PRESTO
        rad_to_deg = 57.29577951308232087679815481410517033240547246656
        sol = 299792458.0
        if self.dish_diameter == -1:
            return -1
        return 0.5 * 1.2 * sol / (freq * 1e6) / \
            self.dish_diameter * rad_to_deg * 3600.0


    def get_string(self):
        """ str method overload """
        fmt = 'Observatory: {}, short name: {}, telescope_id: {}, dish diameter: {}' \
            + ', (X, Y, Z): ({}, {}, {})'
        return fmt.format(self.telescope_name, self.telescope_name_short,
                          self.telescope_id, self.dish_diameter,
                          self.xyz_coords[0], self.xyz_coords[1], self.xyz_coords[2])


if __name__ == "__main__":
    obs = Observatory(telescope_id=6)
    print(obs.get_string())



from __future__ import absolute_import

try:
    from . import waterfall
    from .waterfall import Waterfall
    from .guppi import GuppiRaw
    from . import utils
    from . import fil2h5
    from . import h52fil
    from . import h5diag
    from . import bl_scrunch
    from . import calcload
    from . import rawhdr
    from . import stax
    from . import stix
    from . import match_fils
    from blimpy.io import file_wrapper
except:
    print("Warning: At least one utility could not be imported!")

from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution('blimpy').version
except DistributionNotFound:
    __version__ = '0.0.0 - please install via pip/setup.py'



from blimpy.io import sigproc
import blimpy as bl
from tests.data import voyager_fil, voyager_h5
import numpy as np
import os


def test_sigproc_is_fil():
    """ Check that the is_fil function works """

    assert sigproc.is_filterbank(voyager_h5) is False
    assert sigproc.is_filterbank(voyager_fil) is True


def test_sigproc_generate_headers():
    """ Test if you can generate headers OK from files """
    a = bl.Waterfall(voyager_h5)
    b = bl.Waterfall(voyager_fil)
    sigproc.generate_sigproc_header(a)
    sigproc.generate_sigproc_header(b)

def test_fil_write():
    try:
        a = bl.Waterfall(voyager_h5)
        b = bl.Waterfall(voyager_fil)

        a.write_to_fil('test.fil')
        b.write_to_fil('test2.fil')

        c = bl.Waterfall('test.fil')
        d = bl.Waterfall('test2.fil')

        for key in a.header.keys():
            if key != 'DIMENSION_LABELS':
                assert a.header[key] == c.header[key]
                assert key in c.header.keys()
                assert a.header[key] == d.header[key]
                assert key in d.header.keys()

        assert np.allclose(a.data, c.data)
        assert np.allclose(a.data, d.data)
    except AssertionError:
        print(key, a.header[key], b.header[key], c.header[key], d.header[key])
        raise

    finally:
        os.remove('test.fil')
        os.remove('test2.fil')

if __name__ == "__main__":
    test_sigproc_is_fil()
    test_sigproc_generate_headers()
    test_fil_write()



#!/usr/bin/env python
"""
    Script to dice data to course channel level. From BL FIL of HDF5 files, and outputs HDF5 with '_diced' appended to the file name.

    ..author: Greg Hellbourg (gregory.hellbourg@berkeley.edu)

    March 2018
"""


from .waterfall import Waterfall
import argparse
import math
import sys

#------
# Logging set up
import logging
logger = logging.getLogger(__name__)

level_log = logging.INFO

if level_log == logging.INFO:
    stream = sys.stdout
    format = '%(name)-15s %(levelname)-8s %(message)s'
else:
    stream =  sys.stderr
    format = '%%(relativeCreated)5d (name)-15s %(levelname)-8s %(message)s'

logging.basicConfig(format=format,stream=stream,level = level_log)
#------

def cmd_tool(args=None):
    """ Dices (extracts frequency range) hdf5 or fil files to new file.

    optional arguments:
      -h, --help            show this help message and exit
      -f IN_FNAME, --input_filename IN_FNAME
                            Name of file to write from (HDF5 or FIL)
      -b F_START            Start frequency in MHz
      -e F_STOP             Stop frequency in MHz
      -x OUT_FORMAT, --output_file OUT_FORMAT
                            Output file format [.h5 or .fil].
      -o OUT_FNAME, --output_filename OUT_FNAME
                            Ouput file name to write (to HDF5 or FIL).
      -l MAX_LOAD           Maximum data limit to load.
    """

    parser = argparse.ArgumentParser(description='Dices (extracts frequency range)  hdf5 or fil files and writes to hdf5 or fil.')
    parser.add_argument('-f', '--input_filename', action='store', default=None, dest='in_fname', type=str, help='Name of file to write from (HDF5 or FIL)')
    parser.add_argument('-b', action='store', default=None, dest='f_start', type=float, help='Start frequency in MHz')
    parser.add_argument('-e', action='store', default=None, dest='f_stop', type=float, help='Stop frequency in MHz')
    parser.add_argument('-x', '--output_file', action='store', default=None, dest='out_format', type=str, help='Output file format [.h5 or .fil].')
    parser.add_argument('-o', '--output_filename', action='store', default=None, dest='out_fname', type=str, help='Ouput file name to write (to HDF5 or FIL).')
    parser.add_argument('-l', action='store', default=None, dest='max_load', type=float,help='Maximum data limit to load.')

    if args is None:
        args = sys.argv[1:]

        if len(sys.argv) == 1:
            logger.error('Indicate file name and start and stop frequencies')
            sys.exit()

    args = parser.parse_args(args)

    if args.in_fname is None:
            logger.error('Need to indicate input file name')
            sys.exit()

    if args.out_fname is None:
        if (args.out_format is None) or (args.out_format == 'h5'):
            if args.in_fname[len(args.in_fname)-4:] == '.fil':
                args.out_fname = args.in_fname
                args.out_fname = args.out_fname.replace('.fil','_diced.h5')
            elif args.in_fname[len(args.in_fname)-3:] == '.h5':
                args.out_fname = args.in_fname
                args.out_fname = args.out_fname.replace('.h5','_diced.h5')
            else:
                logger.error('Input file not recognized')
                sys.exit()
        elif args.out_format == 'fil':
            if args.in_fname[len(args.in_fname)-4:] == '.fil':
                args.out_fname = args.in_fname
                args.out_fname = args.out_fname.replace('.fil','_diced.fil')
            elif args.in_fname[len(args.in_fname)-3:] == '.h5':
                args.out_fname = args.in_fname
                args.out_fname = args.out_fname.replace('.h5','_diced.fil')
            else:
                logger.error('input file not recognized.')
                sys.exit()

        else:
            logger.error('Must indicate either output file name or valid output file extension.')
            sys.exit()

    elif (args.out_fname[len(args.out_fname)-4:] == '.fil') and (args.out_format == 'h5'):
        logger.error('Output file extension does not match output file name')
        sys.exit()

    elif (args.out_fname[len(args.out_fname)-3:] == '.h5') and (args.out_format == 'fil'):
        logger.error('Output file extension does not match output file name.')
        sys.exit()

    if (args.out_fname[len(args.out_fname)-3:] != '.h5') and (args.out_fname[len(args.out_fname)-4:] != '.fil'):
        logger.error('Indicate output file name with extension, or simply output file extension.')
        sys.exit()

    if args.f_start == None and args.f_stop == None:
        logger.error('Please give either start and/or end frequencies. Otherwise use fil2h5 or h52fil functions.')
        sys.exit()

    #Read start frequency and bandwidth from data set
    file_big = Waterfall(args.in_fname, max_load = args.max_load)
    f_min_file = file_big.header['fch1']
    f_max_file = file_big.header['fch1'] + file_big.header['nchans'] * file_big.header['foff']

    if args.f_start == None:
        logger.warning('Lower frequency not given, setting to ' + str(f_min_file) + ' MHz to match file.')

    if args.f_stop == None:
        logger.warning('Higher frequency not given, setting to ' + str(f_max_file) + ' MHz to match file.')


    if f_max_file < f_min_file:
            f_max_file,f_min_file = f_min_file,f_max_file

    FreqBWFile = f_max_file-f_min_file
    stdDF = FreqBWFile / float(file_big.calc_n_coarse_chan())   #stdDF = 2.9296875

    if args.f_stop < args.f_start:
            args.f_stop,args.f_start = args.f_start,args.f_stop

    if args.f_start < f_max_file and args.f_start > f_min_file and args.f_stop > f_max_file:
            args.f_stop = f_max_file
            logger.warning('Higher frequency set to ' + str(f_max_file) + ' MHz to match file.')

    if args.f_stop < f_max_file and args.f_stop > f_min_file and args.f_start < f_min_file:
            args.f_start = f_min_file
            logger.warning('Lower frequency set to ' + str(f_min_file) + ' MHz to match file.')

    if args.f_start < f_min_file and args.f_stop > f_max_file:
            args.f_start = f_min_file
            args.f_stop = f_max_file
            logger.warning('Lower frequency set to ' + str(f_min_file) + ' MHz and higher frequency set to ' + str(f_max_file) + ' MHz to match file.')
            # print '\nindicated frequencies include file frequency span - no need to dice\n'
            # sys.exit()

    if min(args.f_start,args.f_stop) < f_min_file or max(args.f_start,args.f_stop) > f_max_file:
            logger.error('Bandwidth to extract must be within ' + str(f_min_file) + ' MHz and ' + str(f_max_file) + ' MHz.')
            sys.exit()

    # calculate real coarse channel begin and end freqs
    f_start_real = math.floor((min(args.f_start,args.f_stop) - f_min_file)/stdDF)*stdDF + f_min_file
    f_stop_real = f_max_file - math.floor((f_max_file - max(args.f_start,args.f_stop))/stdDF)*stdDF

    # print "true start frequency is " + str(f_start_real)
    # print "true stop frequency is " + str(f_stop_real)

    logger.info('Writing to ' + args.out_fname)
    logger.info('Extacting from ' + str(f_start_real) + ' MHz to ' + str(f_stop_real) + ' MHz.')

    # create waterfall object
    file_small = Waterfall(args.in_fname, f_start = f_start_real, f_stop = f_stop_real, max_load = args.max_load)

    # write waterfall object
    if args.out_fname[len(args.out_fname)-4:] == '.fil':
        file_small.write_to_fil(args.out_fname)
    elif args.out_fname[len(args.out_fname)-3:] == '.h5':
        file_small.write_to_hdf5(args.out_fname)
    else:
        logger.error('Error in output file creation : verify output file name and extension.')
        sys.exit()


if __name__ == "__main__":
    cmd_tool()



"""
# test_heavy.py

"""

import blimpy as bl
from tests.data import voyager_fil, voyager_h5


def test_max_data_array_size():
    fw = bl.Waterfall(voyager_fil, max_load=0.001)
    fw = bl.Waterfall(voyager_h5,  max_load=0.001)

if __name__ == "__main__":
    test_max_data_array_size()



r"""
Make waterfall plots of a large file.
"""

import os
import sys
from os.path import dirname, abspath, isdir
import gc
from argparse import ArgumentParser
from PIL import Image
import logging
logger_name = "stix"
logger = logging.getLogger(logger_name)
logger.setLevel(logging.INFO)

# Plotting packages import
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("agg")

# Blimpy imports
from blimpy import Waterfall
from blimpy.plotting import plot_waterfall


def image_stitch(orientation, chunk_count, png_collection, path_saved_png):
    r""" Stitch together multiple PNGs into one"""

    logger.info("Stitching together the images, orientation is {}".format(orientation))

    # Empty array to store loaded images.
    img_collection = []

    # Loop to load images.
    for ii in range(chunk_count):
        img = Image.open(png_collection[ii])
        if ii == 0:
            base_width, base_height = img.size
            logger.info("Base (first) image file pixel width = {}, height = {}"
                        .format(base_width, base_height))
        width, height = img.size
        if width != base_width or height != base_height:
            logger.warning("Skipping {} because pixel width = {}, height = {}, doesn't match the base!"
                           .format(png_collection[ii], width, height))
            continue
        img_collection.append(img)

    # Get dimensions of the last image (same as the rest).

    # Creating the template for the stitched image.
    if orientation == "h":
        new_img = Image.new("RGBA", (chunk_count * width, height))
    else:
        new_img = Image.new("RGBA", (width, chunk_count * height))

    # Initalising the template before pasting the images.
    init_width = 0
    init_height = 0

    # Pasting the images in sequence.
    counter = 0
    for img in img_collection:
        box = (init_width, init_height, init_width + width, init_height + height)
        try:
            new_img.paste(img, box)
        except ValueError:
            logger.error("stix: *** Oops, Image paste error, init_width={}, init_height={}, width={}, height={}"
                         .format(init_width, init_height, width, height))
            sys.exit(86)
        counter += 1
        if orientation == "h":
            init_width += width
        else:
            init_height += height

    # Save the output, re-sized.
    if orientation == "h":
        new_img = new_img.resize((chunk_count * width, height))
    else:
        new_img = new_img.resize((width, chunk_count * height))
    new_img.save(path_saved_png)
    logger.info("Concatenated {} images to {}".format(counter, path_saved_png))


def sort2(x, y):
    r""" Return lowest value, highest value"""
    if y < x:
        return y, x
    return x, y


def make_waterfall_plots(input_file, chunk_count, plot_dir, width, height, dpi, source_name=None):
    r"""
    Make waterfall plots of a given huge-ish file.
​
    Parameters
    ----------
    input_file : str
        Path of Filterbank or HDF5 input file to plot in a stacked mode.
    chunk_count : int
        The number of chunks to divide the entire bandwidth into.
    plot_dir : str
        Directory for storing the PNG files.
    """

    # Get directory path for storing PNG file
    if plot_dir is None:
        dirpath = dirname(abspath(input_file)) + "/"
    else:
        if not isdir(plot_dir):
            os.mkdir(plot_dir)
        dirpath = plot_dir

    # Calculate frequency boundary variables.
    wf1 = Waterfall(input_file, load_data=False)
    nf = wf1.n_channels_in_file
    if nf % chunk_count != 0:
        msg = "Number of frequency chunks ({}) does not evenly divide the number of frequencies ({})!" \
              .format(chunk_count, nf)
        logger.warning(msg)
    fchunk_f_start = wf1.file_header["fch1"]
    fchunk_size = int(nf / chunk_count)
    fchunk_f_offset = fchunk_size * wf1.file_header["foff"]
    fchunk_f_stop = fchunk_f_start + (fchunk_size - 1) * wf1.file_header["foff"]

    # Produce the source_name to be used in the image title.
    if source_name is None:
        source_name = wf1.header["source_name"]
        if source_name.upper() == "UNKNOWN":
            source_name = wf1.header["rawdatafile"].replace(".0000.raw", "")

    # Begin PNG file collection.
    png_collection = []

    # Generate a plot/PNG for each frequency chunk.
    logger.info("Image width = {}, height = {}, dpi = {}"
                .format(width, height, dpi))
    for ii in range(0, chunk_count):

        ii_lowest, ii_highest = sort2(fchunk_f_start, fchunk_f_stop)

        # read in data
        wf = Waterfall(input_file, f_start=ii_lowest, f_stop=ii_highest)

        # Validate frequency range.
        logger.debug("stix: Processing chunk {} of {}, frequency lowest={}, highest={}"
                    .format(ii, chunk_count - 1, ii_lowest, ii_highest))

        # Make plot for this frequency chunk with plot_waterfall().
        wf.header["source_name"] = source_name + " chunk " + str(ii + 1) + " of " + str(chunk_count)
        plt.figure(wf.header["source_name"], figsize=(width, height), dpi=dpi)
        plot_waterfall(wf, f_start=ii_lowest, f_stop=ii_highest)

        # Save the figures.
        path_png = dirpath + source_name + "_chunk_{}".format(ii + 1) + ".png"
        png_collection.append(path_png)
        plt.savefig(path_png, dpi=dpi)
        logger.info("Saved plot: {}".format(path_png))

        # Delete Waterfall object.
        del wf
        gc.collect()
        plt.close("all")

        # Prepare for next chunk.
        fchunk_f_start += fchunk_f_offset
        fchunk_f_stop += fchunk_f_offset

    return png_collection, source_name


def cmd_tool(args=None):
    r"""Coomand line parser"""
    parser = ArgumentParser(description="Make waterfall plots from a single file.")
    parser.add_argument("input_file", type=str, default=None, help="Input file to plot")
    parser.add_argument("chunk_count", type=int, default=0, help="Count of same-sized frequency chunks")
    parser.add_argument("--plot_dir", "-p", type=str, default=".", help="Directory to receive the plot (.png).  Default: current directory.")
    parser.add_argument("--source_name", "-n", type=str, default=None, help="Source name to override the header field.  Default: None.")
    parser.add_argument("--stitch", "-s", type=str, default="n", choices=["n", "v", "h"],
                        help="Stitch files: n(no stitching), v(vertical), or h(horizontal).  Default: n.")
    parser.add_argument("--dpi", "-d", type=int, default=200,
                        help="Single image file dots per inch.  Default: 200.")
    parser.add_argument("--width", "-w", type=float, default=10,
                        help="Single image file image width in inches.  Default: 10.0.")
    parser.add_argument("--height", "-t", type=float, default=8,
                        help="Single image file image height in inches.  Default: 8.0.")

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)
    if args.input_file is None or args.chunk_count < 1:
        os.system("stix -h")
        sys.exit(0)
    if args.dpi < 50:
        logger.error("stix: --dpi must be at least 50 but I saw: {}!".format(args.dpi))
        os.system("stix -h")
        sys.exit(86)
    if args.width < 6:
        logger.error("stix: --width must be at least 6 but I saw: {}!".format(args.width))
        os.system("stix -h")
        sys.exit(86)
    if args.height < 5:
        logger.error("stix: --height must be at least 5 but I saw: {}!".format(args.height))
        os.system("stix -h")
        sys.exit(86)

    # Make the plots.
    logger.info("Begin")
    args.plot_dir += "/"
    args.input_file = os.path.abspath(args.input_file)
    png_collection, source_name = make_waterfall_plots(args.input_file,
                                                       args.chunk_count,
                                                       args.plot_dir,
                                                       args.width,
                                                       args.height,
                                                       args.dpi,
                                                       source_name=args.source_name)
    if args.stitch != "n":
        path_saved_png = args.plot_dir + source_name + ".png"
        image_stitch(args.stitch, args.chunk_count, png_collection, path_saved_png)
        for file in png_collection:
            os.remove(file)

    logger.info("End")


if __name__ == "__main__":
    cmd_tool()



import blimpy as bl
import numpy as np
from pprint import pprint
import pytest
from tests.data import voyager_fil, voyager_h5
from blimpy.plotting.config import plt


def compare_filterbank_fil_to_h5():
    """ Load Voyager dataset and test that both fil and hdf5 readers return same headers and data """

    print("Loading FIL and HDF5 data with Waterfall()..."),
    a = bl.Waterfall(voyager_h5)
    b = bl.Waterfall(voyager_fil)
    print("OK")

    print("Reading headers..")
    print("\nHDF5 file header:")
    pprint(a.header)
    print("\nFIL file header:")
    pprint(b.header)
    print("Headers are loading OK")

    print("\nChecking header values match..."),
    for key in b.header.keys():
        assert b.header[key] == a.header[key]
    print("OK")

    print("Checking datatype matches..."),
    assert a.data.dtype == b.data.dtype
    print("OK")

    print("Checking data matches..."),
    assert np.allclose(a.data, b.data)
    assert a.data.dtype == b.data.dtype
    print("OK")


def test_plotting_doesnt_cause_exceptions():
    """ Try running the plotting routines. They should not raise expections even without X windows """
    a = bl.Waterfall(voyager_h5)
    b = bl.Waterfall(voyager_fil)

    a.plot_all()
    a.plot_kurtosis()
    a.plot_spectrum()
    a.plot_spectrum_min_max()
    a.plot_waterfall()
    a.plot_time_series()
    plt.clf() # Fix issue #140

    b.plot_all()
    b.plot_kurtosis()
    b.plot_spectrum()
    b.plot_spectrum_min_max()
    b.plot_waterfall()
    b.plot_time_series()
    plt.clf()


def test_cmdtool():
    with pytest.raises(SystemExit):
        bl.waterfall.cmd_tool(args=[])


if __name__ == "__main__":
    compare_filterbank_fil_to_h5()
    test_plotting_doesnt_cause_exceptions()

    test_cmdtool()



from os.path import dirname
from blimpy.stax import cmd_line
from tests.data import voyager_fil, voyager_h5


def test_stax():
    dir = dirname(voyager_h5)
    args = [voyager_fil, voyager_h5, "-plot_dir", dir]
    cmd_line(args)
    args = [voyager_fil, voyager_h5, "-plot_dir", dir, "-f_start", "8419", "-f_stop", "8420"]
    cmd_line(args)


if __name__ == "__main__":
    test_stax()



"""
setup.py -- setup script for use of packages.
"""
from setuptools import setup, find_packages

__version__ = '2.0.23'

with open("README.md", "r") as fh:
    long_description = fh.read()

# create entry points
# see http://astropy.readthedocs.org/en/latest/development/scripts.html
entry_points = {
    'console_scripts' : [
        'watutil = blimpy.waterfall:cmd_tool',
        'rawutil = blimpy.guppi:cmd_tool',
        'fil2h5 = blimpy.fil2h5:cmd_tool',
        'h52fil = blimpy.h52fil:cmd_tool',
        'h5diag = blimpy.h5diag:cmd_tool',
        'bl_scrunch = blimpy.bl_scrunch:cmd_tool',
        'matchfils = blimpy.match_fils:cmd_tool',
        'bldice = blimpy.dice:cmd_tool',
        'calcload = blimpy.calcload:cmd_tool',
        'rawhdr = blimpy.rawhdr:cmd_tool',
        'stax = blimpy.stax:cmd_line'
     ]
}

with open("requirements.txt", "r") as fh:
    install_requires = fh.readlines()

extras_require = {
      'full': [
          'pyslalib',
      ]
}

setup(name='blimpy',
      version=__version__,
      description='Python utilities for Breakthrough Listen SETI observations',
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='BSD',
      install_requires=install_requires,
      url='https://github.com/ucberkeleyseti/blimpy',
      author='Danny Price, Emilio Enriquez, Yuhong Chen, Mark Siebert, and BL contributors',
      author_email='dancpr@berkeley.edu',
      entry_points=entry_points,
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Natural Language :: English',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3.7',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Topic :: Scientific/Engineering :: Astronomy',
      ],
      setup_requires=['pytest-runner'],
      tests_require=['pytest', 'pyslalib'],
      test_suite="blimpytests",
)



#!/usr/bin/env python
"""
# guppi.py

A python file handler for guppi RAW files from the GBT.

The guppi raw format consists of a FITS-like header, followed by a block of data,
and repeated over and over until the end of the file.
"""
from __future__ import division
import os
import sys
import numpy as np
from astropy.coordinates import Angle

from .utils import unpack, rebin

PYTHON3 = sys.version_info >= (3, 0)

# Check if $DISPLAY is set (for handling plotting on remote machines with no X-forwarding)
if 'DISPLAY' in os.environ.keys():
    try:
        import matplotlib
    except ImportError:
        pass
    import pylab as plt
else:
    import matplotlib
    matplotlib.use('Agg')
    import pylab as plt

###
# Config values
###

MAX_PLT_POINTS = 65536 * 4  # Max number of points in matplotlib plot
MAX_IMSHOW_POINTS = (8192, 4096)  # Max number of points in imshow plot
MAX_DATA_ARRAY_SIZE = 1024 * 1024 * 1024  # Max size of data array to load into memory


class EndOfFileError(Exception):
    pass


class GuppiRaw():
    """ Python class for reading Guppi raw files

    Args:
        filename (str): name of the .raw file to open

    Optional args:
        n_blocks (int): if number of blocks to read is known, set it here.
                        This saves seeking through the file to check how many
                        integrations there are in the file.
    """

    def __init__(self, filename, n_blocks=None):
        self.filename = filename
        if PYTHON3:
            self.file_obj = open(filename, 'rb')
        else:
            self.file_obj = open(filename, 'rb')
        self.filesize = os.path.getsize(filename)

        if not n_blocks:
            self.n_blocks = self.find_n_data_blocks()

        else:
            self.n_blocks = n_blocks

        self._d = np.zeros(1, dtype='complex64')
        self._d_x = np.zeros(1, dtype='int8')
        self._d_y = np.zeros(1, dtype='int8')
        self.data_gen = None

    def __enter__(self):
        """
        reopen the file each time a `with` block is entered
        :return:
        """
        if PYTHON3:
            self.file_obj = open(self.filename, 'rb')
        else:
            self.file_obj = open(self.filename, 'rb')
        self.filesize = os.path.getsize(self.filename)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """
        closes the file after `with` block has exited
        :param exception_type:
        :param exception_value:
        :param traceback:
        :return:
        """
        self.file_obj.close()

    def __repr__(self):
        return "<GuppiRaw file handler for %s>" % self.filename

    def read_header(self):
        """ Read next header (multiple headers in file)

        Returns:
            (header, data_idx) - a dictionary of keyword:value header data and
            also the byte index of where the corresponding data block resides.
        """
        start_idx = self.file_obj.tell()
        key, val = '', ''

        header_dict = {}
        keep_reading = True

        try:
            while keep_reading:
                if start_idx + 80 > self.filesize:
                    keep_reading = False
                    raise EndOfFileError("End Of Data File")
                line = self.file_obj.read(80)
                if PYTHON3:
                    line = line.decode("utf-8")
                # print line
                if line.startswith('END'):
                    keep_reading = False
                    break

                key, val = line.split('=')
                key, val = key.strip(), val.strip()

                if "'" in val:
                    # Items in quotes are strings
                    val = str(val.strip("'").strip())
                elif "." in val:
                    # Items with periods are floats (if not a string)
                    val = float(val)
                else:
                    # Otherwise it's an integer
                    val = int(val)

                header_dict[key] = val
        except ValueError:
            print("CURRENT LINE: ", line)
            print("BLOCK START IDX: ", start_idx)
            print("FILE SIZE: ", self.filesize)
            print("NEXT 512 BYTES: \n")
            print(self.file_obj.read(512))
            raise

        data_idx = self.file_obj.tell()

        # Seek past padding if DIRECTIO is being used
        if "DIRECTIO" in header_dict.keys():
            if int(header_dict["DIRECTIO"]) == 1:
                if data_idx % 512:
                    data_idx += (512 - data_idx % 512)

        self.file_obj.seek(start_idx)
        return header_dict, data_idx

    def read_first_header(self):
        """ Read first header in file

        Returns:
            header (dict): keyword:value pairs of header metadata
        """
        self.file_obj.seek(0)
        header_dict, unused_pos = self.read_header()
        self.file_obj.seek(0)
        return header_dict

    def read_next_data_block_shape(self, header=None):
        """
        Calculate the shape of the next data block.  Use provided header
        instead of reading header if provided.

        Args:
            header (dict) - keyword:value pairs of header metadata read from
            current block

        Returns:
            dshape (tuple) - shape of the corresponding data block
        """
        if header is None:
            header, unused_data_idx = self.read_header()
        n_chan = int(header['OBSNCHAN'])
        n_pol = int(header['NPOL'])
        n_bit = int(header['NBITS'])
        n_samples = int(int(header['BLOCSIZE']) / (n_chan * n_pol * (n_bit / 8)))

        is_chanmaj = False
        if 'CHANMAJ' in header.keys():
            if int(header['CHANMAJ']) == 1:
                is_chanmaj = True
        if is_chanmaj:
            dshape = (int(n_samples), n_chan, n_pol)
        else:
            dshape = (n_chan, int(n_samples), n_pol)
        return dshape

    def get_data(self):
        """
        returns a generator object that reads data a block at a time;
        the generator prints "File depleted" and returns nothing when all data in the file has been read.
        :return:
        """
        with self as gr:
            while True:
                try:
                    yield gr.generator_read_next_data_block_int8()
                except EndOfFileError as e:
                    print("\nFile depleted")
                    raise StopIteration

    def read_next_data_block_int8(self):
        """
        Instantiates a new generator as self.data_gen if there wasn't one already
        Calls next() on the generator once and returns the value
        Handles generator depletion
        :return: header, data_x, data_y
        """
        if not self.data_gen:
            self.data_gen = self.get_data()
        try:
            header, data_x, data_y = next(self.data_gen)
        except StopIteration:
            self.data_gen = None
            return None, None, None
        self._d_x, self._d_y = data_x, data_y
        return header, np.copy(self._d_x), np.copy(self._d_y)

    def generator_read_next_data_block_int8(self):
        """ Read the next block of data and its header

        Returns: (header, data)
            header (dict): dictionary of header metadata
            data (np.array): Numpy array of data, converted into to complex64.

        """
        header, data_idx = self.read_header()
        self.file_obj.seek(data_idx)

        # Read data and reshape

        n_chan = int(header['OBSNCHAN'])
        n_pol = int(header['NPOL'])
        n_bit = int(header['NBITS'])
        n_samples = int(int(header['BLOCSIZE']) / (n_chan * n_pol * (n_bit / 8)))

        d = np.fromfile(self.file_obj, count=header['BLOCSIZE'], dtype='int8')

        # Handle 2-bit and 4-bit data
        if n_bit != 8:
            d = unpack(d, n_bit)

        d = d.reshape((n_chan, n_samples, n_pol))  # Real, imag

        if self._d_x.shape != d[..., 0:2].shape:
            self._d_x = np.ascontiguousarray(np.zeros(d[..., 0:2].shape, dtype='int8'))
            self._d_y = np.ascontiguousarray(np.zeros(d[..., 2:4].shape, dtype='int8'))

        self._d_x[:] = d[..., 0:2]
        self._d_y[:] = d[..., 2:4]
        return header, self._d_x, self._d_y

    def read_next_data_block_int8_2x(self):
        """ Read the next block of data and its header

        Returns: (header, data)
            header (dict): dictionary of header metadata
            data (np.array): Numpy array of data, converted into to complex64.
        TODO: Deprecate?
        """
        header, data_idx = self.read_header()
        self.file_obj.seek(data_idx)

        # Read data and reshape

        n_chan = int(header['OBSNCHAN'])
        n_pol = int(header['NPOL'])
        n_bit = int(header['NBITS'])
        n_samples = int(int(header['BLOCSIZE']) / (n_chan * n_pol * (n_bit / 8)))

        d = np.fromfile(self.file_obj, count=header['BLOCSIZE'], dtype='int8')

        header, data_idx = self.read_header()
        self.file_obj.seek(data_idx)
        d2 = np.fromfile(self.file_obj, count=header['BLOCSIZE'], dtype='int8')

        # Handle 2-bit and 4-bit data
        if n_bit != 8:
            d = unpack(d, n_bit)

        d = d.reshape((n_chan, n_samples, n_pol))  # Real, imag
        d2 = d2.reshape((n_chan, n_samples, n_pol))
        d = np.concatenate((d, d2), axis=1)
        print(d.shape)

        if self._d_x.shape != (n_chan, n_samples * 2, n_pol):
            self._d_x = np.ascontiguousarray(np.zeros(d[..., 0:2].shape, dtype='int8'))
            self._d_y = np.ascontiguousarray(np.zeros(d[..., 2:4].shape, dtype='int8'))

        self._d_x[:] = d[..., 0:2]
        self._d_y[:] = d[..., 2:4]
        return header, self._d_x, self._d_y

    def read_next_data_block(self):
        """ Read the next block of data and its header

        Returns: (header, data)
            header (dict): dictionary of header metadata
            data (np.array): Numpy array of data, converted into to complex64.
        """
        header, dx, dy = self.read_next_data_block_int8()
        d = np.append(dx, dy, axis=2)

        if self._d.shape != d.shape:
            self._d = np.zeros(d.shape, dtype='float32')

        self._d[:] = d

        return header, self._d[:].view('complex64')

    def find_n_data_blocks(self):
        """ Seek through the file to find how many data blocks there are in the file

        Returns:
            n_blocks (int): number of data blocks in the file
        """
        self.file_obj.seek(0)
        header0, data_idx0 = self.read_header()

        self.file_obj.seek(data_idx0)
        self.file_obj.seek(int(header0['BLOCSIZE']), 1)
        n_blocks = 1
        end_found = False
        while not end_found:
            try:
                header, data_idx = self.read_header()
                self.file_obj.seek(data_idx)
                self.file_obj.seek(header['BLOCSIZE'], 1)
                n_blocks += 1
            except EndOfFileError:
                end_found = True
                break

        self.file_obj.seek(0)
        return n_blocks

    def reset_index(self):
        """ Return file_obj seek to start of file """
        self.file_obj.seek(0)

    def print_stats(self):
        """ Compute some basic stats on the next block of data """

        unused_header, data = self.read_next_data_block()
        data = data.view('float32')

        print("AVG: %2.3f" % data.mean())
        print("STD: %2.3f" % data.std())
        print("MAX: %2.3f" % data.max())
        print("MIN: %2.3f" % data.min())

    def plot_histogram(self, filename=None, flag_show=True):
        """ Plot a histogram of data values

        Args:
            filename (str): Name out output filename. If not set, file will not be saved to disk.

        TODO: Move into plotting/
        """
        unused_header, data = self.read_next_data_block()
        data = data.view('float32')

        plt.figure("Histogram")
        plt.hist(data.flatten(), 65, facecolor='#cc0000')
        if filename:
            plt.savefig(filename)
        if flag_show:
            plt.show()

    def plot_spectrum(self, filename=None, plot_db=True, flag_show=True):
        """ Do a (slow) numpy FFT and take power of data

        Args:
            filename (str): Name out output filename. If not set, file will not be saved to disk.
            plot_db (bool): If True, will plot in dB scale, otherwise linear.

        TODO: Move into plotting/
        """
        unused_header, data = self.read_next_data_block()

        print("Computing FFT...")
        d_xx_fft = np.abs(np.fft.fft(data[..., 0]))
        d_xx_fft = d_xx_fft.flatten()

        # Rebin to max number of points
        dec_fac_x = 1
        if d_xx_fft.shape[0] > MAX_PLT_POINTS:
            dec_fac_x = int(d_xx_fft.shape[0] / MAX_PLT_POINTS)

        d_xx_fft = rebin(d_xx_fft, dec_fac_x)

        print("Plotting...")
        if plot_db:
            plt.plot(10 * np.log10(d_xx_fft))
            plt.ylabel("Power [dB]")
        else:
            plt.plot(d_xx_fft)
            plt.ylabel("Power")
            plt.xlabel("Channel")
        plt.title(self.filename)
        if filename:
            plt.savefig(filename)
        if flag_show:
            plt.show()

    def generate_filterbank_header(self, nchans=1, ):
        """ Generate a blimpy header dictionary

        This function is useful for generating a default header so the raw data
        can be saved into a filterbank file.

        Args:
            nchans (int): Number of fine channels in filterbank header.

        TODO: Deprecate or move to sigproc.py?
        """
        gp_head = self.read_first_header()
        fb_head = {}

        telescope_str = gp_head.get("TELESCOP", "unknown")
        if telescope_str == "unknown":
            print("\n*** WARNING: Missing TELESCOP element in raw header.  Proceeding with a value of 'unknown' ***\n")
        if telescope_str.upper() in ('GBT', 'GREENBANK'):
            fb_head["telescope_id"] = 6
        elif telescope_str.upper() in ('PKS', 'PARKES'):
            fb_head["telescope_id"] = 4
        else:
            fb_head["telescope_id"] = 0

        # Using .get() method allows us to fill in default values if not present
        fb_head["source_name"] = gp_head.get("SRC_NAME", "unknown")
        fb_head["az_start"] = gp_head.get("AZ", 0)
        fb_head["za_start"] = gp_head.get("ZA", 0)

        fb_head["src_raj"] = Angle(str(gp_head.get("RA", 0.0)) + "hr")
        fb_head["src_dej"] = Angle(str(gp_head.get("DEC", 0.0)) + "deg")
        fb_head["rawdatafile"] = self.filename

        # hardcoded
        fb_head["machine_id"] = 42
        fb_head["data_type"] = 1  # blio datatype
        fb_head["barycentric"] = 0
        fb_head["pulsarcentric"] = 0
        fb_head["nbits"] = gp_head.get("NBITS", 32)

        # TODO - compute these values. Need to figure out the correct calcs
        fb_head["tstart"] = 0.0
        fb_head["tsamp"] = 1.0
        fb_head["fch1"] = 0.0
        fb_head["foff"] = gp_head.get("OBSBW", 187.5) / nchans

        # Need to be updated based on output specs
        fb_head["nchans"] = nchans
        fb_head["nifs"] = 1
        fb_head["nbeams"] = 1

        return fb_head


def cmd_tool(args=None):
    """ Command line tool for plotting and viewing info on GUPPI Raw files """

    from argparse import ArgumentParser

    parser = ArgumentParser(description="Command line utility quick look at GUPPI Raw files. ")

    parser.add_argument('filename', type=str, help='Name of file to read')
    parser.add_argument('-o', dest='outdir', type=str, default='./', help='output directory for PNG files')
    args = parser.parse_args()

    r = GuppiRaw(args.filename)

    r.print_stats()
    bname = os.path.splitext(os.path.basename(args.filename))[0]
    bname = os.path.join(args.outdir, bname)
    r.plot_histogram(filename="%s_hist.png" % bname)
    r.plot_spectrum(filename="%s_spec.png" % bname)


if __name__ == "__main__":
    cmd_tool()



r'''Reader for HDF5 files'''
import os

import h5py
import numpy as np
from astropy.coordinates import Angle

from blimpy.io.base_reader import Reader, logger, GIGA


def examine_h5(h5):
    if "CLASS" in h5.attrs:
        classstr = h5.attrs["CLASS"]
    else:
        raise ValueError("HDF5 CLASS attribute missing")
    if not classstr in ["FILTERBANK", b"FILTERBANK"]:
        raise ValueError("Expected HDF5 CLASS attribute to be 'FILTERBANK' but saw '{}'".format(classstr))
    if not "VERSION" in h5.attrs:
        raise ValueError("HDF5 VERSION attribute missing")
    if not "data" in h5:
        raise ValueError("HDF5 data matrix missing")
    if h5["data"].ndim != 3:
        raise ValueError("Expected HDF5 data.ndim to be 3 but saw '{}'".format(h5["data"].ndim))


class H5Reader(Reader):
    """ This class handles .h5 files.
    """

    def __init__(self, filename, f_start=None, f_stop=None, t_start=None, t_stop=None, load_data=True, max_load=None):
        """ Constructor.

        Args:
            filename (str): file path of HDF5 file.
            f_start (float): start frequency, in MHz
            f_stop (float): stop frequency, in MHz
            t_start (int): start time bin
            t_stop (int): stop time bin
            max_load (float): memory limit in gigabytes
        """
        super(H5Reader, self).__init__()

        if not os.path.isfile(filename):
            raise IOError("Not a file: {}".format(filename))
        if not h5py.is_hdf5(filename):
            raise IOError("Not an HDF5 file: {}".format(filename))

        #These values may be modified once code for multi_beam and multi_stokes observations are possible.
        self.freq_axis = 2
        self.time_axis = 0
        self.beam_axis = 1  # Place holder
        self.stokes_axis = 4  # Place holder

        self.filename = filename
        self.filestat = os.stat(filename)
        self.filesize = self.filestat.st_size/(1024.0**2)
        self.load_data = load_data
        self.h5 = h5py.File(self.filename, mode='r')
        examine_h5(self.h5)
        self.read_header()
        self.file_size_bytes = os.path.getsize(self.filename)  # In bytes
        self.n_ints_in_file = self.h5["data"].shape[self.time_axis] #
        self.n_channels_in_file = self.h5["data"].shape[self.freq_axis] #
        self.n_beams_in_file = self.header['nifs'] #Placeholder for future development.
        self.n_pols_in_file = 1 #Placeholder for future development.
        self._n_bytes = int(self.header['nbits'] / 8)  #number of bytes per digit.
        self._d_type = self._setup_dtype()
        self.file_shape = (self.n_ints_in_file,self.n_beams_in_file,self.n_channels_in_file)

        if self.header['foff'] < 0:
            self.f_end = self.header['fch1']
            self.f_begin = self.f_end + self.n_channels_in_file*self.header['foff']
        else:
            self.f_begin = self.header['fch1']
            self.f_end = self.f_begin + self.n_channels_in_file*self.header['foff']

        self.t_begin = 0
        self.t_end = self.n_ints_in_file

        #Taking care all the frequencies are assigned correctly.
        self._setup_selection_range(f_start=f_start, f_stop=f_stop, t_start=t_start, t_stop=t_stop, init=True)
        #Convert input frequencies into what their corresponding channel number would be.
        self._setup_chans()
        #Update frequencies ranges from channel number.
        self._setup_freqs()

        #Applying data size limit to load.
        if max_load is not None and max_load > 0:
            self.max_data_array_size = max_load * GIGA

        if self.file_size_bytes > self.max_data_array_size:
            self.large_file = True
        else:
            self.large_file = False

        if self.load_data:
            if self.large_file:
                # Only checking the selection, if the file is too large.
                if self.f_start or self.f_stop or self.t_start or self.t_stop:
                    if self.isheavy():
                        self.warn_memory("Selection", self._calc_selection_size())
                        self._init_empty_selection()
                    else:
                        self.read_data()
                else:
                    self.warn_memory("File", self.file_size_bytes)
                    self._init_empty_selection()
            else:
                self.read_data()
        else:
            logger.debug("Skipping loading data ...")
            self._init_empty_selection()


    def read_header(self):
        """ Read header and return a Python dictionary of key:value pairs
        """

        self.header = {}

        for key, val in self.h5['data'].attrs.items():
            #if six.PY3:
            #    key = bytes(key, 'ascii')
            if isinstance(val, bytes):
                val = val.decode('ascii')
            if key == 'src_raj':
                self.header[key] = Angle(val, unit='hr')
            elif key == 'src_dej':
                self.header[key] = Angle(val, unit='deg')
            else:
                self.header[key] = val

        return self.header

    def _find_blob_start(self, blob_dim, n_blob):
        """Find first blob from selection.
        """

        #Convert input frequencies into what their corresponding channel number would be.
        self._setup_chans()

        #Check which is the blob time offset
        blob_time_start = self.t_start + blob_dim[self.time_axis]*n_blob

        #Check which is the blob frequency offset (in channels)
        blob_freq_start = self.chan_start_idx + (blob_dim[self.freq_axis]*n_blob)%self.selection_shape[self.freq_axis]

        blob_start = np.array([blob_time_start, 0, blob_freq_start])

        return blob_start

    def read_data(self, f_start=None, f_stop=None,t_start=None, t_stop=None):
        """ Read data
        """

        self._setup_selection_range(f_start=f_start, f_stop=f_stop, t_start=t_start, t_stop=t_stop)

        #check if selection is small enough.
        if self.isheavy():
            self.warn_memory("Selection", self._calc_selection_size())
            self.data = np.array([0], dtype=self._d_type)
            return

        #Convert input frequencies into what their corresponding channel number would be.
        self._setup_chans()
        #Update frequencies ranges from channel number.
        self._setup_freqs()

        self.data = self.h5["data"][self.t_start:self.t_stop,:,self.chan_start_idx:self.chan_stop_idx]

    def read_blob(self,blob_dim,n_blob=0):
        """Read blob from a selection.
        """

        n_blobs = self.calc_n_blobs(blob_dim)
        if n_blob > n_blobs or n_blob < 0:
            raise ValueError('Please provide correct n_blob value. Given %i, but max values is %i'%(n_blob,n_blobs))

        #This prevents issues when the last blob is smaller than the others in time
        if blob_dim[self.time_axis]*(n_blob+1) > self.selection_shape[self.time_axis]:
            updated_blob_dim = (self.selection_shape[self.time_axis] - blob_dim[self.time_axis]*n_blob, 1, blob_dim[self.freq_axis])
        else:
            updated_blob_dim = [int(i) for i in blob_dim]

        blob_start = self._find_blob_start(blob_dim, n_blob)
        blob_end = blob_start + np.array(updated_blob_dim)

        blob = self.h5["data"][int(blob_start[self.time_axis]):int(blob_end[self.time_axis]),
                               :,
                               int(blob_start[self.freq_axis]):int(blob_end[self.freq_axis])
                               ]

#         if self.header['foff'] < 0:
#             blob = blob[:,:,::-1]

        return blob



#!/usr/bin/env python
"""
# waterfall.py

Python class and command line utility for reading and plotting waterfall files.

This provides a class, Waterfall(), which can be used to read a blimpy file (.fil or .h5):


    fil = Waterfall("test_psr.fil")
    print(fil.header)
    print(fil.data.shape)
    print(fil.freqs)

    plt.figure()
    fil.plot_spectrum(t=0)
    plt.show()


TODO: check the file seek logic works correctly for multiple IFs

"""

import sys
import os
import numpy as np
import six

from blimpy.io import file_wrapper as fw
from .plotting import *

from astropy.time import Time
from astropy import units as u

# YANKED 2021-03-05: from numba import jit

#------
# Logging set up
import logging
logger = logging.getLogger(__name__)

level_log = logging.INFO

if level_log == logging.INFO:
    stream = sys.stdout
    fmt = '%(name)-15s %(levelname)-8s %(message)s'
else:
    stream =  sys.stderr
    fmt = '%%(relativeCreated)5d (name)-15s %(levelname)-8s %(message)s'

logging.basicConfig(format=fmt, stream=stream, level=level_log)


#import pdb #pdb.set_trace()

MAX_BLOB_MB = 1024


###
# Main blimpy class
###

class Waterfall():
    """ Class for loading and writing blimpy data (.fil, .h5) """

    """ Get the frequency array for this Waterfall object"""    
    get_freqs = lambda self: np.arange(0, self.header['nchans'], 1, dtype=float) * self.header['foff'] + self.header['fch1']

    def __repr__(self):
        return "Waterfall data: %s" % self.filename

    def __init__(self, filename=None, f_start=None, f_stop=None, t_start=None, t_stop=None,
                 load_data=True, max_load=None, header_dict=None, data_array=None):
        """ Class for loading and plotting blimpy data.

        This class parses the blimpy file and stores the header and data
        as objects:
            fb = Waterfall('filename_here.fil')
            fb.header        # blimpy header, as a dictionary
            fb.data          # blimpy data, as a numpy array

        Args:
            filename (str): filename of blimpy file.
            f_start (float): start frequency in MHz
            f_stop (float): stop frequency in MHz
            t_start (int): start integration ID
            t_stop (int): stop integration ID
            load_data (bool): load data. If set to False, only header will be read.
            max_load (float): maximum data to load in GB.
            header_dict (dict): Create blimpy from header dictionary + data array
            data_array (np.array): Create blimpy from header dict + data array
        """

##EE        super(Waterfall, self).__init__()

        if filename:
            self.filename = filename
            self.ext = os.path.splitext(filename)[-1].lower()
            self.container = fw.open_file(filename, f_start=f_start, f_stop=f_stop, t_start=t_start, t_stop=t_stop,
                                          load_data=load_data, max_load=max_load)
            self.file_header = self.container.header
            self.header = self.file_header
            self.n_ints_in_file = self.container.n_ints_in_file
            self.file_shape = self.container.file_shape
            self.file_size_bytes = self.container.file_size_bytes
            self.selection_shape = self.container.selection_shape
            self.n_channels_in_file = self.container.n_channels_in_file

            # These values will be modified once code for multi_beam and multi_stokes observations are possible.
            self.freq_axis = 2
            self.time_axis = 0
            self.beam_axis = 1  # Place holder
            self.stokes_axis = 4  # Place holder

            self.logger = logger

            self.__load_data()

        elif header_dict is not None and data_array is not None:
            self.filename = ''
            self.header = header_dict
            self.data = data_array
            self.n_ints_in_file = 0
            self._setup_freqs()

        else:
            self.filename = ''

        # Attach methods
        self.plot_spectrum         = six.create_bound_method(plot_spectrum, self)
        self.plot_waterfall        = six.create_bound_method(plot_waterfall, self)
        self.plot_kurtosis         = six.create_bound_method(plot_kurtosis, self)
        self.plot_time_series      = six.create_bound_method(plot_time_series, self)
        self.plot_all              = six.create_bound_method(plot_all, self)
        self.plot_spectrum_min_max = six.create_bound_method(plot_spectrum_min_max, self)
    
    def __load_data(self):
        """ Helper for loading data from a container. Should not be called manually. """

        self.data = self.container.data

    def read_data(self, f_start=None, f_stop=None,t_start=None, t_stop=None):
        """ Reads data selection if small enough.

        Args:
            f_start (float): Start frequency in MHz
            f_stop (float): Stop frequency in MHz
            t_start (int): Integer time index to start at
            t_stop (int): Integer time index to stop at

        Data is loaded into self.data (nothing is returned)
        """

        self.container.read_data(f_start=f_start, f_stop=f_stop,t_start=t_start, t_stop=t_stop)

        self.__load_data()

    def _update_header(self):
        """ Updates the header information from the original file to the selection. """

        #Updating frequency of first channel from selection
        if self.header['foff'] < 0:
            self.header['fch1'] = self.container.f_stop
        else:
            self.header['fch1'] = self.container.f_start

        #Updating number of fine channels.
        self.header['nchans'] = self.container.selection_shape[self.freq_axis]

        #Updating time stamp for first time bin from selection
        self.header['tstart'] = self.container.populate_timestamps(update_header=True)

    def info(self):
        """ Print header information and other derived information. """

        print("\n--- File Info ---")

        for key, val in self.file_header.items():
            if key == 'src_raj':
                val = val.to_string(unit=u.hour, sep=':')
            if key == 'src_dej':
                val = val.to_string(unit=u.deg, sep=':')
            if key in ('foff', 'fch1'):
                val *= u.MHz
            if key == 'tstart':
                print("%16s : %32s" % ("tstart (ISOT)", Time(val, format='mjd').isot))
                key = "tstart (MJD)"
            print("%16s : %32s" % (key, val))

        print("\n%16s : %32s" % ("Num ints in file", self.n_ints_in_file))
        print("%16s : %32s" % ("File shape", self.file_shape))
        print("--- Selection Info ---")
        print("%16s : %32s" % ("Data selection shape", self.selection_shape))
        if self.header['foff'] < 0:
            minfreq = self.container.f_start - self.header['foff']
            maxfreq = self.container.f_stop
        else:
            minfreq = self.container.f_start
            maxfreq = self.container.f_stop - self.header['foff']
        print("%16s : %32s" % ("Minimum freq (MHz)", minfreq))
        print("%16s : %32s" % ("Maximum freq (MHz)", maxfreq))


    def _get_blob_dimensions(self, chunk_dim):
        """ Sets the blob dimensions, trying to read around 1024 MiB at a time.
            This is assuming a chunk is about 1 MiB.

            Notes:
                A 'blob' is the max size that will be read into memory at once.
                A 'chunk' is a HDF5 concept to do with efficient read access, see
                https://portal.hdfgroup.org/display/HDF5/Chunking+in+HDF5

            Args:
                chunk_dim (array of ints): Shape of chunk, e.g. (1024, 1, 768)

            Returns size of blob (array of ints).
        """

        #Taking the size into consideration, but avoiding having multiple blobs within a single time bin.
        if self.selection_shape[self.freq_axis] > chunk_dim[self.freq_axis]*MAX_BLOB_MB:
            freq_axis_size = self.selection_shape[self.freq_axis]
#             while freq_axis_size > chunk_dim[self.freq_axis]*MAX_BLOB_MB:
#                 freq_axis_size /= 2
            time_axis_size = 1
        else:
            freq_axis_size = self.selection_shape[self.freq_axis]
            time_axis_size = np.min([chunk_dim[self.time_axis] * MAX_BLOB_MB * chunk_dim[self.freq_axis] / freq_axis_size, self.selection_shape[self.time_axis]])

        blob_dim = (int(time_axis_size), 1, freq_axis_size)

        return blob_dim

    def _get_chunk_dimensions(self):
        """ Sets the chunking dimensions depending on the file type.

            Notes: A 'chunk' is a HDF5 concept to do with efficient read access, see
            https://portal.hdfgroup.org/display/HDF5/Chunking+in+HDF5

            Returns chunk dimensions, e.g. (2048, 1, 512)
        """

        #Usually '.0000.' is in self.filename
        if np.abs(self.header['foff']) < 1e-5:
            logger.info('Detecting high frequency resolution data.')
            chunk_dim = (1,1,1048576) #1048576 is the number of channels in a coarse channel.
            return chunk_dim

        #Usually '.0001.' is in self.filename
        if np.abs(self.header['tsamp']) < 1e-3:
            logger.info('Detecting high time resolution data.')
            chunk_dim = (2048,1,512) #512 is the total number of channels per single band (ie. blc00)
            return chunk_dim

        #Usually '.0002.' is in self.filename
        if np.abs(self.header['foff']) < 1e-2 and np.abs(self.header['foff'])  >= 1e-5:
            logger.info('Detecting intermediate frequency and time resolution data.')
            chunk_dim = (10,1,65536)  #65536 is the total number of channels per single band (ie. blc00)
            return chunk_dim

        logger.warning('File format not known. Will use minimum chunking. NOT OPTIMAL.')
        chunk_dim = (1,1,512)
        return chunk_dim

    def calc_n_coarse_chan(self, chan_bw=None):
        """ This makes an attempt to calculate the number of coarse channels in a given freq selection.

            Note:
                This is unlikely to work on non-Breakthrough Listen data, as a-priori knowledge of
                the digitizer system is required.

            Returns n_coarse_chan (int), number of coarse channels
        """

        n_coarse_chan = self.container.calc_n_coarse_chan(chan_bw)

        return n_coarse_chan

    def grab_data(self, f_start=None, f_stop=None,t_start=None, t_stop=None, if_id=0):
        """ Extract a portion of data by frequency range.

        Args:
            f_start (float): start frequency in MHz
            f_stop (float): stop frequency in MHz
            if_id (int): IF input identification (req. when multiple IFs in file)

        Returns:
            (freqs, data) (np.arrays): frequency axis in MHz and data subset
        """

        if self.container.isheavy():
            raise Exception("Waterfall.grab_data: Large data array was not loaded!  Try instantiating Waterfall with max_load.")

        try:
            self.freqs = self.container.populate_freqs()
        except Exception as ex:
            raise Exception("Waterfall.grab_data: Cannot allocate frequency array") from ex

        try:
            self.timestamps = self.container.populate_timestamps()
        except Exception as ex:
            raise Exception("Waterfall.grab_data: Cannot allocate timestamps array") from ex

        if f_start is None:
            f_start = self.freqs[0]
        if f_stop is None:
            f_stop = self.freqs[-1]

        try:
            i0 = np.argmin(np.abs(self.freqs - f_start))
            i1 = np.argmin(np.abs(self.freqs - f_stop))

            if i0 < i1:
                plot_f    = self.freqs[i0:i1 + 1]
                plot_data = np.squeeze(self.data[t_start:t_stop, if_id, i0:i1 + 1])
            else:
                plot_f    = self.freqs[i1:i0 + 1]
                plot_data = np.squeeze(self.data[t_start:t_stop, if_id, i1:i0 + 1])
        except:
            raise Exception("Waterfall.grab_data: Too much data requested")

        return plot_f, plot_data

    def write_to_fil(self, filename_out, *args, **kwargs):
        from blimpy.io import write_to_fil
        write_to_fil(self, filename_out)

    def write_to_hdf5(self, filename_out, *args, **kwargs):
        from blimpy.io import write_to_hdf5
        write_to_hdf5(self, filename_out, *args, **kwargs)

    def blank_dc(self, n_coarse_chan):
        """ Blank DC bins in coarse channels.

        Removes the DC spike in centre of coarse channel bins.

        Note: currently only works if entire file is read
        """

        if n_coarse_chan < 1:
            logger.warning('Coarse channel number < 1, unable to blank DC bin.')
            return None

        if not n_coarse_chan % int(n_coarse_chan) == 0:
            logger.warning('Selection does not contain an integer number of coarse channels, unable to blank DC bin.')
            return None

        n_coarse_chan = int(n_coarse_chan)

        n_chan = self.data.shape[-1]
        n_chan_per_coarse = int(n_chan / n_coarse_chan) # ratio of fine channels to coarse channels

        mid_chan = int(n_chan_per_coarse / 2)

        def parse(data, n_coarse_chan, n_chan_per_coarse, mid_chan):
            for ii in range(n_coarse_chan):
                ss = ii*n_chan_per_coarse
                w_slice = data[..., ss+mid_chan+5:ss+mid_chan+10]
                chtest = w_slice.shape[-1]
                # If we are nearing the end of the fine channel frequency array,
                # do not do anymore.  See issue #212.
                if chtest < 5:
                    break
                data[..., ss+mid_chan] = np.median(w_slice)
        
        parse(self.data, n_coarse_chan, n_chan_per_coarse, mid_chan)

    def calibrate_band_pass_N1(self):
        """ One way to calibrate the band pass is to take the median value
            for every frequency fine channel, and divide by it.

            sets data = data / band_pass
        """

        band_pass = np.median(self.data.squeeze(),axis=0)
        self.data = self.data/band_pass


def cmd_tool(args=None):
    """ Command line tool for plotting and viewing info on blimpy files """


    from argparse import ArgumentParser

    parser = ArgumentParser(description="Command line utility for reading and plotting blimpy files.")

    parser.add_argument('filename', type=str,
                        help='Name of file to read')
    parser.add_argument('-p', action='store',  default='a', dest='what_to_plot', type=str,
                        help='Show: "w" waterfall (freq vs. time) plot; "s" integrated spectrum plot; \
                        "t" for time series; "mm" for spectrum including min max; "k" for kurtosis; \
                        "a" for all available plots and information; and "ank" for all but kurtosis.')
    parser.add_argument('-b', action='store', default=None, dest='f_start', type=float,
                        help='Start frequency (begin), in MHz')
    parser.add_argument('-e', action='store', default=None, dest='f_stop', type=float,
                        help='Stop frequency (end), in MHz')
    parser.add_argument('-B', action='store', default=None, dest='t_start', type=int,
                        help='Start integration (begin, inclusive) ID ')
    parser.add_argument('-E', action='store', default=None, dest='t_stop', type=int,
                        help='Stop integration (end, exclusive) ID')
    parser.add_argument('-i', action='store_true', default=False, dest='info_only',
                        help='Show info only')
    parser.add_argument('-a', action='store_true', default=False, dest='average',
                       help='average along time axis (plot spectrum only)')
    parser.add_argument('-s', action='store', default='', dest='plt_filename', type=str,
                       help='save plot graphic to file (give filename as argument)')
    parser.add_argument('-S', action='store_true', default=False, dest='save_only',
                       help='Turn off plotting of data and only save to file.')
    parser.add_argument('-D', action='store_false', default=False, dest='blank_dc',
                       help='Use to blank DC bin.')
    parser.add_argument('-H', action='store_true', default=False, dest='to_hdf5',
                       help='Write file to hdf5 format.')
    parser.add_argument('-F', action='store_true', default=False, dest='to_fil',
                       help='Write file to .fil format.')
    parser.add_argument('-o', action='store', default=None, dest='filename_out', type=str,
                        help='Filename output (if not provided, the name will be the same but with appropriate extension).')
    parser.add_argument('-l', action='store', default=None, dest='max_load', type=float,
                        help='Maximum data limit to load. Default:1GB')

    if args is None:
        args = sys.argv[1:]

    parse_args = parser.parse_args(args)

    # Open blimpy data
    filename = parse_args.filename
    load_data = not parse_args.info_only
    info_only = parse_args.info_only
    filename_out = parse_args.filename_out

    fil = Waterfall(filename, f_start=parse_args.f_start, f_stop=parse_args.f_stop, t_start=parse_args.t_start, t_stop=parse_args.t_stop, load_data=load_data, max_load=parse_args.max_load)
    fil.info()

    #Check the size of selection.
    if fil.container.isheavy() or parse_args.to_hdf5 or parse_args.to_fil:
        info_only = True

    # And if we want to plot data, then plot data.
    if not info_only:
        from .plotting.config import plt

        print('')

        if parse_args.blank_dc:
            logger.info("Blanking DC bin")
            n_coarse_chan = fil.calc_n_coarse_chan()
            fil.blank_dc(n_coarse_chan)

        if parse_args.what_to_plot == "w":
            plt.figure("waterfall", figsize=(8, 6))
            fil.plot_waterfall(f_start=parse_args.f_start, f_stop=parse_args.f_stop)
        elif parse_args.what_to_plot == "s":
            plt.figure("Spectrum", figsize=(8, 6))
            fil.plot_spectrum(logged=True, f_start=parse_args.f_start, f_stop=parse_args.f_stop, t='all')
        elif parse_args.what_to_plot == "mm":
            plt.figure("min max", figsize=(8, 6))
            fil.plot_spectrum_min_max(logged=True, f_start=parse_args.f_start, f_stop=parse_args.f_stop, t='all')
        elif parse_args.what_to_plot == "k":
            plt.figure("kurtosis", figsize=(8, 6))
            fil.plot_kurtosis(f_start=parse_args.f_start, f_stop=parse_args.f_stop)
        elif parse_args.what_to_plot == "t":
            plt.figure("Time Series", figsize=(8, 6))
            fil.plot_time_series(f_start=parse_args.f_start, f_stop=parse_args.f_stop,orientation='h')
        elif parse_args.what_to_plot == "a":
            plt.figure("Multiple diagnostic plots", figsize=(12, 9),facecolor='white')
            fil.plot_all(logged=True, f_start=parse_args.f_start, f_stop=parse_args.f_stop, t='all')
        elif parse_args.what_to_plot == "ank":
            plt.figure("Multiple diagnostic plots", figsize=(12, 9),facecolor='white')
            fil.plot_all(logged=True, f_start=parse_args.f_start, f_stop=parse_args.f_stop, t='all',kutosis=False)

        if parse_args.plt_filename != '':
            plt.savefig(parse_args.plt_filename)

        if not parse_args.save_only:
            if 'DISPLAY' in os.environ.keys():
                plt.show()
            else:
                logger.warning("No $DISPLAY available.")

    else:

        fileroot = os.path.splitext(filename)[0]

        if parse_args.to_hdf5 and parse_args.to_fil:
            # If this is the last statement anyway,
            # and consequently, an output file is not generated,
            # why do we not raise an error?
            raise ValueError('Either provide to_hdf5 or to_fil, but not both.')

        if parse_args.to_hdf5:
            if not filename_out:
                filename_out = fileroot + '.h5'

            logger.info('Writing file : %s'% filename_out)
            fil.write_to_hdf5(filename_out)
            logger.info('File written.')

        elif parse_args.to_fil:
            if not filename_out:
                filename_out = fileroot + '.fil'

            logger.info('Writing file : %s'% filename_out)
            fil.write_to_fil(filename_out)
            logger.info('File written.')


if __name__ == "__main__":
    cmd_tool()



import numpy as np

import logging
import psutil
import sys

logger = logging.getLogger(__name__)

level_log = logging.INFO

if level_log == logging.INFO:
    stream = sys.stdout
    lformat = '%(name)-15s %(levelname)-8s %(message)s'
else:
    stream =  sys.stderr
    lformat = '%%(relativeCreated)5d (name)-15s %(levelname)-8s %(message)s'

logging.basicConfig(format=lformat, stream=stream, level=level_log)

# Convenient for memory calculations
GIGA = 1024 ** 3

# Threshold size for high resolution data
HIRES_THRESHOLD = 2**20


class Reader(object):
    """ Basic reader object """

    def __init__(self):

        self.t_begin = 0
        self.t_end   = 0

        # Calculate the max data array size from available memory
        self.available_memory = psutil.virtual_memory().available
        logger.debug("Reader __init__ available_memory={}".format(self.available_memory))
        if self.available_memory > GIGA:
            self.max_data_array_size = self.available_memory - GIGA
        else:
            self.max_data_array_size = self.available_memory
            logger.warning("Very low on memory, only {:.2f} MB available for use."
                           .format(float(self.available_memory) / 1e6))
        logger.debug("Reader __init__ max_data_array_size={}".format(self.max_data_array_size))

    def _setup_selection_range(self, f_start=None, f_stop=None, t_start=None, t_stop=None, init=False):
        """Making sure the selection if time and frequency are within the file limits.

        Args:
            init (bool): If call during __init__
        """


        def _dump_parms():
            wstr = "f_start={}, f_stop={}, t_start={}, t_stop={}, init={}" \
                  .format(f_start, f_stop, t_start, t_stop, init)
            logger.warning(wstr)


        # This avoids resetting values
        if init is True:
            if t_start is None:
                t_start = self.t_begin
            if t_stop is None:
                t_stop = self.t_end
            if f_start is None:
                f_start = self.f_begin
            if f_stop is None:
                f_stop = self.f_end
        else:
            if f_start is None:
                f_start = self.f_start
            if f_stop is None:
                f_stop = self.f_stop
            if t_start is None:
                t_start = self.t_start
            if t_stop is None:
                t_stop = self.t_stop

        # By now, all values start/stop are populated.

        if t_stop >= 0 and t_start >= 0 and t_stop < t_start:
            t_stop, t_start = t_start,t_stop
            logger.warning('Given t_stop < t_start, assuming reversed values.')
        if f_stop and f_start and f_stop < f_start:
            f_stop, f_start = f_start,f_stop
            logger.warning('Given f_stop < f_start, assuming reversed values.')

        if t_start >= self.t_begin and t_start < self.t_end:
            self.t_start = int(t_start)
        else:
            if init is False or t_start != None:
                logger.warning('Setting t_start = %f, since t_start not given or not valid.'%self.t_begin)
                _dump_parms()
            self.t_start = self.t_begin

        if t_stop <= self.t_end  and t_stop > self.t_begin:
            self.t_stop = int(t_stop)
        else:
            if init is False or t_stop:
                logger.warning('Setting t_stop = %f, since t_stop not given or not valid.'%self.t_end)
                _dump_parms()
            self.t_stop = self.t_end

        if f_start >= self.f_begin and f_start < self.f_end:
            self.f_start = f_start
        else:
            if init is False or f_start:
                logger.warning('Setting f_start = %f, since f_start not given or not valid.'%self.f_begin)
                _dump_parms()
            self.f_start = self.f_begin

        if f_stop <= self.f_end and f_stop > self.f_begin:
            self.f_stop = f_stop
        else:
            if init is False or f_stop:
                logger.warning('Setting f_stop = %f, since f_stop not given or not valid.'%self.f_end)
                _dump_parms()
            self.f_stop = self.f_end

        # Now we have setup bounds, we can calculate shape of selection
        self.selection_shape = self._calc_selection_shape()

    def _init_empty_selection(self):
        """
        """

        self.data = np.array([0],dtype=self._d_type)

    def _setup_dtype(self):
        """Calculating dtype
        """

        #Set up the data type
        if self._n_bytes  == 4:
            return 'float32'
        elif self._n_bytes  == 2:
            return 'uint16'
        elif self._n_bytes  == 1:
            return 'uint8'
        else:
            errmsg = 'Reader._setup_dtype detected invalid/unsupported header data: _n_bytes={}, nbits={}'.format(self._n_bytes, self.header['nbits'])
            logger.error(errmsg)
            raise ValueError(errmsg)

    def _calc_selection_size(self):
        """Calculate size of data of interest.
        """

        #Check to see how many integrations requested
        n_ints = self.t_stop - self.t_start
        #Check to see how many frequency channels requested
        n_chan = (self.f_stop - self.f_start) / abs(self.header['foff'])

        n_bytes  = self._n_bytes
        selection_size = int(n_ints*n_chan*n_bytes)

        return selection_size

    def _calc_selection_shape(self):
        """Calculate shape of data of interest.
        """

        #Check how many integrations requested
        n_ints = int(self.t_stop - self.t_start)
        #Check how many frequency channels requested
        n_chan = int(np.round((self.f_stop - self.f_start) / abs(self.header['foff'])))

        selection_shape = (n_ints,int(self.header['nifs']),n_chan)

        return selection_shape

    def _setup_chans(self):
        """Setup channel borders
        """

        if self.header['foff'] < 0:
            f0 = self.f_end
        else:
            f0 = self.f_begin

        i_start, i_stop = 0, self.n_channels_in_file
        if self.f_start:
            i_start = np.round((self.f_start - f0) / self.header['foff'])
        if self.f_stop:
            i_stop  = np.round((self.f_stop - f0)  / self.header['foff'])

        #calculate closest true index value
        chan_start_idx = int(i_start)
        chan_stop_idx  = int(i_stop)

        if chan_stop_idx < chan_start_idx:
            chan_stop_idx, chan_start_idx = chan_start_idx,chan_stop_idx

        self.chan_start_idx =  chan_start_idx
        self.chan_stop_idx = chan_stop_idx

    def _setup_freqs(self):
        """Updating frequency borders from channel values
        """

        if self.header['foff'] > 0:
            self.f_start = self.f_begin + self.chan_start_idx*abs(self.header['foff'])
            self.f_stop = self.f_begin + self.chan_stop_idx*abs(self.header['foff'])
        else:
            self.f_start = self.f_end - self.chan_stop_idx*abs(self.header['foff'])
            self.f_stop = self.f_end - self.chan_start_idx*abs(self.header['foff'])

    def populate_timestamps(self,update_header=False):
        """  Populate time axis.
            IF update_header then only return tstart
        """

        #Check to see how many integrations requested
        ii_start, ii_stop = 0, self.n_ints_in_file
        if self.t_start:
            ii_start = self.t_start
        if self.t_stop:
            ii_stop = self.t_stop

        ## Setup time axis
        t0 = self.header['tstart']
        t_delt = self.header['tsamp']

        if update_header:
            timestamps = ii_start * t_delt / 24./60./60. + t0
        else:
            timestamps = np.arange(ii_start, ii_stop) * t_delt / 24./60./60. + t0

        return timestamps

    def populate_freqs(self):
        """
         Populate frequency axis
        """

        if self.header['foff'] < 0:
            f0 = self.f_end
        else:
            f0 = self.f_begin

        self._setup_chans()

        #create freq array
        i_vals = np.arange(self.chan_start_idx, self.chan_stop_idx)
        freqs = self.header['foff'] * i_vals + f0

        return freqs

    def adjust_n_coarse_chan(self, n_coarse_chan, nchans):
        r"""Don't let the calculated n_coarse_chan be < 1
            nor exceed the number of fine channels."""
        if n_coarse_chan < 1:
            errmsg1 = "blimpy:io:base_reader:adjust_n_coarse_chan: n_coarse_chan={}, nchans={}" \
                      .format(n_coarse_chan, nchans)
            logger.warning(errmsg1)
            errmsg2 = "blimpy:io:base_reader:adjust_n_coarse_chan: n_coarse_chan < 1. Replacing that with a value of 64 (SWAG)."
            logger.warning(errmsg2)
            return 64
        if n_coarse_chan > nchans: # exceeds the number of fine channels?
            errmsg1 = "blimpy:io:base_reader:adjust_n_coarse_chan: n_coarse_chan={}, nchans={}" \
                      .format(n_coarse_chan, nchans)
            logger.warning(errmsg1)
            errmsg2 = "blimpy:io:base_reader:adjust_n_coarse_chan: n_coarse_chan > nchans. Replacing that with the value of nchans (SWAG)."
            logger.warning(errmsg2)
            return nchans
        return n_coarse_chan

    def calc_n_coarse_chan(self, chan_bw=None):
        """ This makes an attempt to calculate the number of coarse channels in a given file.

            Note:
                This is unlikely to work on non-Breakthrough Listen data, as a-priori knowledge of
                the digitizer system is required.

        """

        nchans = int(self.header['nchans'])

        # Do we have a file with enough channels that it has coarse channelization?
        if chan_bw is not None:
            bandwidth = abs(self.f_stop - self.f_start)
            n_coarse_chan = bandwidth / chan_bw
            return self.adjust_n_coarse_chan(n_coarse_chan, nchans)

        # High resolution data?
        if nchans >= HIRES_THRESHOLD:
            # Does the common FFT length of 2^20 divide through without a remainder?
            # This should work for most GBT and all Parkes hires data
            if nchans % HIRES_THRESHOLD == 0:
                n_coarse_chan = nchans / float(HIRES_THRESHOLD)
                return self.adjust_n_coarse_chan(n_coarse_chan, nchans)
            # Early GBT data has non-2^N FFT length, check if it is GBT data
            elif self.header['telescope_id'] == 6:
                coarse_chan_bw = 2.9296875
                bandwidth = abs(self.f_stop - self.f_start)
                n_coarse_chan = bandwidth / coarse_chan_bw
                return self.adjust_n_coarse_chan(n_coarse_chan, nchans)
            else:
                errmsg1 = "blimpy:io:base_reader:calc_n_coarse_chan: hires nchans not divisible by 2^20 and not GBT"
                logger.warning(errmsg1)
                errmsg2 = "Setting a value of 64 (SWAG). In turbo_seti, you can specify n_course_chan explicitly."
                logger.info(errmsg2)
                return 64
                
        # Not high resolution data.  GBT?
        if self.header['telescope_id'] == 6:
            #For GBT non-hires data
            coarse_chan_bw = 2.9296875
            bandwidth = abs(self.f_stop - self.f_start)
            n_coarse_chan = bandwidth / coarse_chan_bw
            return self.adjust_n_coarse_chan(n_coarse_chan, nchans)
            
        # Not high resolution data and Not GBT
        else: 
            errmsg1 = "blimpy:io:base_reader:calc_n_coarse_chan: not hires and not GBT. Setting a value of 1"
            logger.warning(errmsg1)
            errmsg2 = "Setting a value of 64 (SWAG). In turbo_seti, you can specify n_course_chan explicitly."
            logger.info(errmsg2)
            return 64

    def calc_n_blobs(self, blob_dim):
        """ Given the blob dimensions, calculate how many fit in the data selection.
        """

        n_blobs = int(np.ceil(1.0 * np.prod(self.selection_shape) / np.prod(blob_dim)))

        return n_blobs

    def warn_memory(self, name, size):
        """ Warn that <name> is larger than our limit for loading things into memory.
        """
        logger.warning(f"{name} size is {size / GIGA:.2f} GB, which exceeds the memory usage limit of {self.max_data_array_size / GIGA} GB. Keeping data on disk.")
    
    def isheavy(self):
        """ Check if the current selection is too large.
        """

        selection_size_bytes = self._calc_selection_size()

        if selection_size_bytes > self.max_data_array_size:
            return True
        else:
            return False



from blimpy import Waterfall
import numpy as np
from astropy import units as u

fwhm_to_sig = 1./(8*np.log(2))**0.5 # a constant


def foldcal(data,tsamp, diode_p=0.04,numsamps=1000,switch=False,inds=False):
    '''
    Returns time-averaged spectra of the ON and OFF measurements in a
    calibrator measurement with flickering noise diode

    Parameters
    ----------
    data : 2D Array object (float)
        2D dynamic spectrum for data (any Stokes parameter) with flickering noise diode.
    tsamp : float
        Sampling time of data in seconds
    diode_p : float
        Period of the flickering noise diode in seconds
    numsamps : int
        Number of samples over which to average noise diode ON and OFF
    switch : boolean
        Use switch=True if the noise diode "skips" turning from OFF to ON once or vice versa
    inds : boolean
        Use inds=True to also return the indexes of the time series where the ND is ON and OFF
    '''

    halfper = diode_p/2.0

    foldt = halfper/tsamp   #number of time samples per diode switch

    onesec = 1/tsamp    #number of time samples in the first second

    #Find diode switches in units of time samples and round down to the nearest int
    ints = np.arange(0,numsamps)
    t_switch = (onesec+ints*foldt)
    t_switch = t_switch.astype('int')

    ONints = np.array(np.reshape(t_switch[:],(numsamps/2,2)))
    ONints[:,0] = ONints[:,0]+1   #Find index ranges of ON time samples

    OFFints = np.array(np.reshape(t_switch[1:-1],(numsamps/2-1,2)))
    OFFints[:,0] = OFFints[:,0]+1   #Find index ranges of OFF time samples

    av_ON = []
    av_OFF = []

    #Average ON and OFF spectra separately with respect to time
    for i in ONints:
        if i[1]!=i[0]:
            av_ON.append(np.sum(data[i[0]:i[1],:,:],axis=0)/(i[1]-i[0]))

    for i in OFFints:
        if i[1]!=i[0]:
            av_OFF.append(np.sum(data[i[0]:i[1],:,:],axis=0)/(i[1]-i[0]))

    #If switch=True, flip the return statement since ON is actually OFF
    if switch:
        if not inds:
            return np.squeeze(np.mean(av_OFF,axis=0)), np.squeeze(np.mean(av_ON,axis=0))
        return np.squeeze(np.mean(av_OFF,axis=0)), np.squeeze(np.mean(av_ON,axis=0)),OFFints,ONints

    if not inds:
        return np.squeeze(np.mean(av_ON,axis=0)), np.squeeze(np.mean(av_OFF,axis=0))
    return np.squeeze(np.mean(av_ON,axis=0)), np.squeeze(np.mean(av_OFF,axis=0)),ONints,OFFints

def integrate_chans(spec, chan_per_coarse):
    '''
    Integrates over each core channel of a given spectrum.
    Important for calibrating data with frequency/time resolution different from noise diode data

    Parameters
    ----------
    spec : 1D Array (float)
        Spectrum (any Stokes parameter) to be integrated
    chan_per_coarse: int
        Number of frequency bins per coarse channel
    '''

    num_coarse = spec.size/chan_per_coarse     #Calculate total number of coarse channels

    #Rearrange spectrum by coarse channel
    spec_shaped = np.array(np.reshape(spec,(num_coarse,chan_per_coarse)))

    #Average over coarse channels
    return np.mean(spec_shaped[:,1:-1],axis=1)

def integrate_calib(name,chan_per_coarse,fullstokes=False,**kwargs):
    '''
    Folds Stokes I noise diode data and integrates along coarse channels

    Parameters
    ----------
    name : str
        Path to noise diode filterbank file
    chan_per_coarse : int
        Number of frequency bins per coarse channel
    fullstokes : boolean
        Use fullstokes=True if data is in IQUV format or just Stokes I, use fullstokes=False if
        it is in cross_pols format
    '''
    #Load data
    obs = Waterfall(name,max_load=150)
    data = obs.data

    #If the data has cross_pols format calculate Stokes I
    if fullstokes==False and data.shape[1]>1:
        data = data[:,0,:]+data[:,1,:]
        data = np.expand_dims(data,axis=1)
    #If the data has IQUV format get Stokes I
    if fullstokes==True:
        data = data[:,0,:]
        data = np.expand_dims(data,axis=1)
    # So, what should happen if fullstokes==False and data.shape[1]<=1 ???

    tsamp = obs.header['tsamp']

    #Calculate ON and OFF values
    OFF,ON = foldcal(data,tsamp,**kwargs)

    #Find ON and OFF spectra by coarse channel
    ON_int = integrate_chans(ON, chan_per_coarse)
    OFF_int = integrate_chans(OFF, chan_per_coarse)

    #If "ON" is actually "OFF" switch them
    if np.sum(ON_int)<np.sum(OFF_int):
        temp = ON_int
        ON_int = OFF_int
        OFF_int = temp

    #Return coarse channel spectrum of OFF and ON
    return OFF_int,ON_int

def get_calfluxes(calflux,calfreq,spec_in,centerfreqs,oneflux):
    '''
    Given properties of the calibrator source, calculate fluxes of the source
    in a particular frequency range

    Parameters
    ----------
    calflux : float
        Known flux of calibrator source at a particular frequency
    calfreq : float
        Frequency where calibrator source has flux calflux (MHz) (see above)
    spec_in : float
        Known power-law spectral index of calibrator source. Use convention flux(frequency) = constant * frequency^(spec_in)
    centerfreqs : 1D Array (float)
        Central frequency values of each coarse channel
    oneflux : boolean
        Use oneflux to choose between calculating the flux for each core channel (False)
        or using one value for the entire frequency range (True)
    '''

    const = calflux/np.power(calfreq,spec_in)
    if oneflux:
        return const*np.power(np.mean(centerfreqs),spec_in)
    return const*np.power(centerfreqs,spec_in)

def get_centerfreqs(freqs,chan_per_coarse):
    '''
    Returns central frequency of each coarse channel

    Parameters
    ----------
    freqs : 1D Array (float)
        Frequency values for each bin of the spectrum
    chan_per_coarse: int
        Number of frequency bins per coarse channel
    '''

    num_coarse = freqs.size/chan_per_coarse
    freqs = np.reshape(freqs,(num_coarse,chan_per_coarse))
    return np.mean(freqs,axis=1)

def f_ratios(calON_obs,calOFF_obs,chan_per_coarse,**kwargs):
    '''
    Calculate f_ON, and f_OFF as defined in van Straten et al. 2012 equations 2 and 3

    Parameters
    ----------
    calON_obs : str
        Path to filterbank file (any format) for observation ON the calibrator source
    calOFF_obs : str
        Path to filterbank file (any format) for observation OFF the calibrator source
    '''
    #Calculate noise diode ON and noise diode OFF spectra (H and L) for both observations
    L_ON,H_ON = integrate_calib(calON_obs,chan_per_coarse,**kwargs)
    L_OFF,H_OFF = integrate_calib(calOFF_obs,chan_per_coarse,**kwargs)

    f_ON = H_ON/L_ON-1
    f_OFF = H_OFF/L_OFF-1

    return f_ON, f_OFF


def diode_spec(calON_obs,calOFF_obs,calflux,calfreq,spec_in,average=True,oneflux=False,**kwargs):
    '''
    Calculate the coarse channel spectrum and system temperature of the noise diode in Jy given two noise diode
    measurements ON and OFF the calibrator source with the same frequency and time resolution

    Parameters
    ----------
    calON_obs : str
        (see f_ratios() above)
    calOFF_obs : str
        (see f_ratios() above)
    calflux : float
        Known flux of calibrator source at a particular frequency
    calfreq : float
        Frequency where calibrator source has flux calflux (see above)
    spec_in : float
        Known power-law spectral index of calibrator source. Use convention flux(frequency) = constant * frequency^(spec_in)
    average : boolean
        Use average=True to return noise diode and Tsys spectra averaged over frequencies
    '''
    #Load frequencies and calculate number of channels per coarse channel
    obs = Waterfall(calON_obs,max_load=150)
    freqs = obs.container.populate_freqs()
    ncoarse = obs.calc_n_coarse_chan()
    nchans = obs.header['nchans']
    chan_per_coarse = nchans/ncoarse

    f_ON, f_OFF = f_ratios(calON_obs,calOFF_obs,chan_per_coarse,**kwargs)

    #Obtain spectrum of the calibrator source for the given frequency range
    centerfreqs = get_centerfreqs(freqs,chan_per_coarse)
    calfluxes = get_calfluxes(calflux,calfreq,spec_in,centerfreqs,oneflux)

    #C_o and Tsys as defined in van Straten et al. 2012
    C_o = calfluxes/(1/f_ON-1/f_OFF)
    Tsys = C_o/f_OFF

    #return coarse channel diode spectrum
    if average:
        return np.mean(C_o),np.mean(Tsys)
    return C_o,Tsys

def get_Tsys(calON_obs, calOFF_obs, calflux, calfreq, spec_in, **kwargs):
    '''
    Returns frequency dependent system temperature given observations on and off a calibrator source

    Parameters
    ----------
    (See diode_spec())
    '''
    obs = Waterfall(calON_obs,max_load=150)
    ncoarse = obs.calc_n_coarse_chan()
    chan_per_coarse = obs.header['nchans']/ncoarse
    freqs = obs.container.populate_freqs()
    cfreqs = get_centerfreqs(freqs,chan_per_coarse)
    S_sys = diode_spec(calON_obs,calOFF_obs,calflux,calfreq,spec_in,average=False,oneflux=False,**kwargs)[1]

    T_sys = Jy_to_Kelvin(S_sys,cfreqs)
    return T_sys,cfreqs

def get_Tsys_nodiode(calON_obs_name, calOFF_obs_name, calflux, calfreq, spec_in):
    '''
    Calculates system temperature from two flux calibrator scans taken without noise diode flickering.
    CURRENTLY ONLY IMPLEMENTED FOR STOKES I DATA.

    Parameters
    ----------
    calON_obs_name : str
        Path to filterbank file for scan ON calibrator target
    calOFF_obs_name : str
        Path to filterbank file for scan OFF calibrator
    calflux : float
        Flux in Jy of the calibrator source
    calfreq : float or int
        Frequency at which calflux was taken (MHz)
    spec_in : float
        Spectral index of this calibrator
    '''
    calON_obs = Waterfall(calON_obs_name,max_load=150)
    calOFF_obs = Waterfall(calOFF_obs_name,max_load=150)
    ncoarse = calON_obs.calc_n_coarse_chan()
    chan_per_coarse = calON_obs.header['nchans']/ncoarse

    ONobs_spec = np.squeeze(np.mean(calON_obs.data,axis=0))
    OFFobs_spec = np.squeeze(np.mean(calOFF_obs.data,axis=0))
    freqs = calON_obs.container.populate_freqs()

    ON_coarse_spec = integrate_chans(ONobs_spec, chan_per_coarse)
    OFF_coarse_spec = integrate_chans(OFFobs_spec, chan_per_coarse)
    cfreqs = get_centerfreqs(freqs,chan_per_coarse)
    cal_fluxes = get_calfluxes(calflux,calfreq,spec_in,cfreqs,oneflux=False)

    S_sys = (OFF_coarse_spec)/(ON_coarse_spec-OFF_coarse_spec)*cal_fluxes

    #Convert to Kelvins
    T_sys,cfreqs = Jy_to_Kelvin(S_sys,cfreqs)
    return T_sys,cfreqs


def Jy_to_Kelvin(flux,freqs):
    '''
    Convert flux spectrum in Jy/beam to temperature units. Frequency inputs in MHz.

    Parameters
    ----------
    flux : 1D Array (float)
        Spectrum over frequency band in Jy
    freqs : 1D Array (float)
        Frequencies (in MHz)
    '''
    beam_FWHM = 755.0/(np.mean(freqs*0.001))*u.arcsec
    beam_area = 2*np.pi*(beam_FWHM*fwhm_to_sig)**2
    freq_MHz = np.mean(freqs)*u.MHz
    equiv = u.brightness_temperature(beam_area,freq_MHz)
    T = (flux*u.Jy).to(u.K,equivalencies=equiv)

    return T,freqs

def calibrate_fluxes(main_obs_name,dio_name,dspec,Tsys,fullstokes=False,**kwargs):
    '''
    Produce calibrated Stokes I for an observation given a noise diode
    measurement on the source and a diode spectrum with the same number of
    coarse channels

    Parameters
    ----------
    main_obs_name : str
        Path to filterbank file containing final data to be calibrated
    dio_name : str
        Path to filterbank file for observation on the target source with flickering noise diode
    dspec : 1D Array (float) or float
        Coarse channel spectrum (or average) of the noise diode in Jy (obtained from diode_spec())
    Tsys : 1D Array (float) or float
        Coarse channel spectrum (or average) of the system temperature in Jy
    fullstokes: boolean
        Use fullstokes=True if data is in IQUV format or just Stokes I, use fullstokes=False if
        it is in cross_pols format
    '''

    #Find folded spectra of the target source with the noise diode ON and OFF
    main_obs = Waterfall(main_obs_name,max_load=150)
    ncoarse = main_obs.calc_n_coarse_chan()
    dio_obs = Waterfall(dio_name,max_load=150)
    dio_chan_per_coarse = dio_obs.header['nchans']/ncoarse
    dOFF,dON = integrate_calib(dio_name,dio_chan_per_coarse,fullstokes,**kwargs)

    #Find Jy/count for each coarse channel using the diode spectrum
    main_dat = main_obs.data
    scale_facs = dspec/(dON-dOFF)
    print(scale_facs)

    nchans = main_obs.header['nchans']
    obs_chan_per_coarse = nchans/ncoarse

    ax0_size = np.size(main_dat,0)
    ax1_size = np.size(main_dat,1)

    #Reshape data array of target observation and multiply coarse channels by the scale factors
    main_dat = np.reshape(main_dat,(ax0_size,ax1_size,ncoarse,obs_chan_per_coarse))
    main_dat = np.swapaxes(main_dat,2,3)

    main_dat = main_dat*scale_facs
    main_dat = main_dat-Tsys
    main_dat = np.swapaxes(main_dat,2,3)
    main_dat = np.reshape(main_dat,(ax0_size,ax1_size,nchans))

    #Write calibrated data to a new filterbank file with ".fluxcal" extension

    main_obs.data = main_dat
    main_obs.write_to_fil(main_obs_name[:-4]+'.fluxcal.fil')
    print('Finished: calibrated product written to ' + main_obs_name[:-4]+'.fluxcal.fil')



from .config import *
from ..utils import rebin, db

def plot_spectrum_min_max(wf, t=0, f_start=None, f_stop=None, logged=False, if_id=0, c=None, **kwargs):
    """ Plot frequency spectrum of a given file

    Args:
        logged (bool): Plot in linear (False) or dB units (True)
        if_id (int): IF identification (if multiple IF signals in file)
        c: color for line
        kwargs: keyword args to be passed to matplotlib plot()
    """
    ax = plt.gca()

    plot_f, plot_data = wf.grab_data(f_start, f_stop, if_id)

    # Using accending frequency for all plots.
    if wf.header['foff'] < 0:
        plot_data = plot_data[..., ::-1]  # Reverse data
        plot_f = plot_f[::-1]

    if logged:
        db_plot_data = db(plot_data[0])
        fig_max = np.nanmax(db_plot_data[db_plot_data != np.inf])
        fig_min = np.nanmin(db_plot_data[db_plot_data != -np.inf])
    else:
        fig_max = plot_data[0].max()
        fig_min = plot_data[0].min()

    print("averaging along time axis...")

    # Since the data has been squeezed, the axis for time goes away if only one bin, causing a bug with axis=1
    if len(plot_data.shape) > 1:
        plot_max = plot_data.max(axis=0)
        plot_min = plot_data.min(axis=0)
        plot_data = plot_data.mean(axis=0)
    else:
        plot_max = plot_data.max()
        plot_min = plot_data.min()
        plot_data = plot_data.mean()

    # Rebin to max number of points
    dec_fac_x = 1
    MAX_PLT_POINTS = 8 * 64  # Low resoluition to see the difference.
    if plot_data.shape[0] > MAX_PLT_POINTS:
        dec_fac_x = int(plot_data.shape[0] / MAX_PLT_POINTS)

    plot_data = rebin(plot_data, dec_fac_x, 1)
    plot_min = rebin(plot_min, dec_fac_x, 1)
    plot_max = rebin(plot_max, dec_fac_x, 1)
    plot_f = rebin(plot_f, dec_fac_x, 1)

    if logged:
        plt.plot(plot_f, db(plot_data), "#333333", label='mean', **kwargs)
        plt.plot(plot_f, db(plot_max), "#e74c3c", label='max', **kwargs)
        plt.plot(plot_f, db(plot_min), '#3b5b92', label='min', **kwargs)
        plt.ylabel("Power [dB]")
    else:
        plt.plot(plot_f, plot_data, "#333333", label='mean', **kwargs)
        plt.plot(plot_f, plot_max, "#e74c3c", label='max', **kwargs)
        plt.plot(plot_f, plot_min, '#3b5b92', label='min', **kwargs)
        plt.ylabel("Power [counts]")
    plt.xlabel("Frequency [MHz]")
    plt.legend()

    try:
        plt.title(wf.header['source_name'])
    except KeyError:
        plt.title(wf.filename)

    plt.xlim(plot_f[0], plot_f[-1])
    if logged:
        try:
            plt.ylim(fig_min - 1, fig_max + 1)
        except ValueError:
            plt.ylim(-10, 20)



"""
# test_voyager_data_load.py

The hard-coded numbers in these tests can be found in the voyager_test_setup.ipynb

"""

import blimpy as bl
import numpy as np
import pylab as plt
from tests.data import voyager_fil, voyager_h5


def test_waterfall_data_load_range_freq():
    fw = bl.Waterfall(voyager_fil, f_start=8419.24, f_stop=8419.35)
    hw = bl.Waterfall(voyager_h5, f_start=8419.24, f_stop=8419.35)

    print(fw.data.shape)
    print(hw.data.shape)
    print(hw.data[0].max(), hw.data[0].argmax())
    print(fw.data[0].max(), fw.data[0].argmax())
    print(hw.data[-1].max(), hw.data[-1].argmax())
    print(fw.data[-1].max(), fw.data[-1].argmax())

    # Assert data is loaded to the same shape and has same values
    assert hw.data.shape == fw.data.shape == (16, 1, 39370)
    assert np.allclose(hw.data, fw.data)

    # Check the Voyager carrier has the known amplitudes at first and last integration
    assert np.allclose(hw.data[0].max(), fw.data[0].max(), 3.09333e+11)
    assert np.allclose(hw.data[-1].max(), fw.data[-1].max(), 2.74257e+11)

    # Check the tone is in the same bin for both
    assert hw.data[0].argmax() == fw.data[0].argmax() == 18959
    assert hw.data[-1].argmax() == fw.data[-1].argmax() == 18996

    # And plot
    plt.figure("VOYAGER DATA LOAD")
    plt.subplot(2,1,1)
    fw.plot_spectrum()

    plt.subplot(2,1,2)
    hw.plot_spectrum()
    plt.tight_layout()
    #plt.clf()

def test_grab_data_works_across_all_fil_h5():

    fw = bl.Waterfall(voyager_fil)
    hw = bl.Waterfall(voyager_h5)
    all_readers = [fw, hw]

    for ii, rr in enumerate(all_readers):
        f, d = rr.grab_data(f_start=8419.29, f_stop=8419.30)
        print(f.shape, d.shape)
        assert f.shape == (3580,)
        assert d.shape == (16, 3580)

    for ii, rr in enumerate(all_readers):
        f, d = rr.grab_data(f_start=8419.29685, f_stop=8419.2971)
        print(f.shape, d.shape)
        assert f.shape == (91,)
        assert d.shape == (16, 91)

if __name__ == "__main__":
    test_waterfall_data_load_range_freq()
    test_grab_data_works_across_all_fil_h5()



from blimpy import Waterfall
from tests.data import voyager_h5, test_ifs_fil


HIRES_THRESHOLD = 2**20


def test_ncc_chan_bw():
    wf = Waterfall(voyager_h5)
    print("test_bcc_chan_bw: telescope_id:", wf.header['telescope_id'])
    print("test_ncc_chan_bw: nchans:", wf.header['nchans'])

    n_coarse_chan = wf.calc_n_coarse_chan(16)
    print("test_ncc_chan_bw: n_coarse_chan [chan_bw=16]:", n_coarse_chan)
    assert n_coarse_chan == 64

    n_coarse_chan = wf.calc_n_coarse_chan(1)
    print("test_ncc_chan_bw: n_coarse_chan [chan_bw=1]:", n_coarse_chan)
    assert n_coarse_chan > 2.9 and n_coarse_chan < 3.0
    
    
def test_ncc_gbt():
    wf = Waterfall(test_ifs_fil)
    wf.header['telescope_id'] = 6
    print("test_ncc_gbt: telescope_id:", wf.header['telescope_id'])
    print("test_ncc_gbt: starting nchans:", wf.header['nchans'])

    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_gbt: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 64

    wf.header['nchans'] = 3 * HIRES_THRESHOLD
    print("\ntest_ncc_gbt: nchans:", wf.header['nchans'])
    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_gbt: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 3

    wf.header['nchans'] = HIRES_THRESHOLD
    print("\ntest_ncc_gbt: nchans:", wf.header['nchans'])
    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_gbt: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 1

    wf.header['nchans'] = HIRES_THRESHOLD - 1
    print("\ntest_ncc_gbt: nchans:", wf.header['nchans'])
    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_gbt: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 64

    wf.header['nchans'] = HIRES_THRESHOLD + 1
    print("\ntest_ncc_gbt: nchans:", wf.header['nchans'])
    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_gbt: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 64


def test_ncc_42():
    wf = Waterfall(test_ifs_fil)
    wf.header['telescope_id'] = 42
    print("test_ncc_42: telescope_id:", wf.header['telescope_id'])
    print("test_ncc_42: starting nchans:", wf.header['nchans'])

    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_42: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 64

    wf.header['nchans'] = 3 * HIRES_THRESHOLD
    print("\ntest_ncc_42: nchans:", wf.header['nchans'])
    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_42: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 3

    wf.header['nchans'] = HIRES_THRESHOLD
    print("\ntest_ncc_42: nchans:", wf.header['nchans'])
    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_42: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 1

    wf.header['nchans'] = HIRES_THRESHOLD - 1
    print("\ntest_ncc_42: nchans:", wf.header['nchans'])
    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_42: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 64

    wf.header['nchans'] = HIRES_THRESHOLD + 1
    print("\ntest_ncc_42: nchans:", wf.header['nchans'])
    n_coarse_chan = wf.calc_n_coarse_chan()
    print("test_ncc_42: n_coarse_chan [chan_bw=None]:", n_coarse_chan)
    assert n_coarse_chan == 64


if __name__ == "__main__":
    wf = Waterfall(test_ifs_fil)
    wf.info()
    test_ncc_chan_bw()
    test_ncc_gbt()
    test_ncc_42()
